from __future__ import annotations

import datetime
import re
from functools import lru_cache
from typing import TYPE_CHECKING, Any

from dissect.eventlog import evtx
from dissect.eventlog.exceptions import MalformedElfChnkException
from flow.record import Record, utils

from dissect.target.exceptions import FilesystemError
from dissect.target.helpers.record import DynamicDescriptor, TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export
from dissect.target.plugins.os.windows.log.evt import WindowsEventlogsMixin

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.target import Target

re_illegal_characters = re.compile(r"[\(\): \.\-#\/\>\<]")


EVTX_GLOB = "*.evtx"


class EvtxPlugin(WindowsEventlogsMixin, Plugin):
    """Plugin for fetching and parsing Windows Eventlog Files (``*.evtx``)."""

    RECORD_NAME = "filesystem/windows/evtx"
    LOGS_DIR_PATH = "%windir%/system32/winevt/logs"

    NEEDLE = b"ElfChnk\x00"
    CHUNK_SIZE = 0x10000

    def __init__(self, target: Target):
        super().__init__(target)
        self._create_event_descriptor = lru_cache(4096)(self._create_event_descriptor)

    @arg("--logs-dir", help="logs directory to scan")
    @arg("--log-file-glob", default=EVTX_GLOB, help="glob pattern to match a log file name")
    @export(record=DynamicDescriptor(["datetime"]))
    def evtx(self, log_file_glob: str = EVTX_GLOB, logs_dir: str | None = None) -> Iterator[DynamicDescriptor]:
        """Return entries from Windows Event log files (``*.evtx``).

        Windows Event log is a detailed record of system, security and application notifications. It can be used to
        diagnose a system or find future issues. Up until Windows XP the extension .evt was used, hereafter ``.evtx``
        became the new standard.

        References:
            - https://www.techtarget.com/searchwindowsserver/definition/Windows-event-log
            - https://serverfault.com/questions/441050/what-are-the-differences-between-windows-evt-and-evtx-log-files

        Yields dynamically created records based on the fields in the event.
        At least contains the following fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The TimeCreated_SystemTime field of the event.
            Provider_Name (string): The Provider_Name field of the event.
            EventID (int): The EventID of the event.
        """

        if logs_dir:
            log_paths = self.get_logs_from_dir(logs_dir, filename_glob=log_file_glob)
        else:
            log_paths = self.get_logs(filename_glob=log_file_glob)

        for entry in log_paths:
            if not entry.exists():
                self.target.log.warning("Event log file does not exist: %s", entry)
                continue

            try:
                entry_data = entry.open()
            except FilesystemError:
                self.target.log.exception("Failed to open event log: %s", entry)
                continue

            self.target.log.info("Processing event log file %s", entry)
            try:
                for event in evtx.Evtx(entry_data):
                    try:
                        yield self._build_record(event, entry)
                    except Exception as e:  # noqa: PERF203
                        self.target.log.warning("Unable to parse event log event %s: %s", event, e)
                        self.target.log.debug("", exc_info=e)
            except EOFError as e:
                self.target.log.warning("Unable to parse event log file %s: %s", entry, e)
                self.target.log.debug("", exc_info=e)

    @export(record=DynamicDescriptor(["datetime"]))
    def scraped_evtx(self) -> Iterator[DynamicDescriptor]:
        """Return EVTX log file records scraped from target disks."""
        yield from self.target.scrape.scrape_chunks_from_disks(
            needle=self.NEEDLE,
            chunk_size=self.CHUNK_SIZE,
            chunk_parser=self._parse_chunk,
        )

    def _parse_chunk(self, needle: bytes, chunk: bytes) -> Iterator[Record]:
        chnk = evtx.ElfChnk(chunk)
        try:
            for event in chnk.read():
                yield self._build_record(event, None)
        except MalformedElfChnkException:
            pass

    def _build_record(self, evtx_record: dict, source: Path | None) -> Record:
        # predictable order of fields in the list is important, since we'll
        # be constructing a record descriptor from it.
        evtx_record_fields = sorted(evtx_record.items())

        record_values = {
            "_target": self.target,
            "source": source,
        }
        record_fields = [
            ("datetime", "ts"),
            ("string", "Provider_Name"),
            ("uint32", "EventID"),
        ]
        unk_fields = 0

        for key, value in evtx_record_fields:
            key = re_illegal_characters.sub("_", key).strip("_")
            value = format_value(value)

            if not key.isprintable():
                self.target.log.warning(
                    "Skipped possibly corrupt field record containing non-printable characters: %s", key
                )
                continue

            if key == "TimeCreated_SystemTime":
                record_values["ts"] = value
                continue

            if isinstance(value, list):
                for idx, v in enumerate(value):
                    k = f"{key}_{idx}"
                    record_values[k] = format_value(v)
                    record_fields.append(("string", k))
            else:
                is_invalid = False
                try:
                    key.encode("utf8")
                except UnicodeEncodeError:
                    is_invalid = True
                if not key or key[0].isdigit() or is_invalid:
                    key = f"unknown_{unk_fields}"
                    unk_fields += 1

                if key == "EventID":
                    value = int(value)

                if key in record_values:
                    key = unique_key(key, record_values)

                record_values[key] = value

                if key in ("Provider_Name", "EventID"):
                    continue

                record_fields.append(("string", key))

        record_fields.append(("path", "source"))

        # tuple conversion here is needed for lru_cache
        desc = self._create_event_descriptor(tuple(record_fields))
        return desc(**record_values)

    def _create_event_descriptor(self, record_fields: list[tuple[str, str]]) -> TargetRecordDescriptor:
        return TargetRecordDescriptor(self.RECORD_NAME, record_fields)


def format_value(value: Any) -> Any:
    if value is None or value == "-":
        return None

    if isinstance(value, evtx.BxmlSub):
        value = value.get()

    if isinstance(value, (datetime.datetime, list)):
        return value

    try:
        return utils.to_str(value)
    except UnicodeDecodeError:
        return repr(value)


def unique_key(key: str, dictionary: dict[str, Any], count: int | None = None) -> str:
    """Return a unique key for a given dict of key value pairs.

    Makes the returned key unique by appending an incrementing integer after the given key name (e.g. ``key_2``).
    Search is case sensitive so provide lower-cased ``key`` and ``dictionary`` arguments if case-insensitiveness
    is desired.
    """
    count = count or 2
    new_key = f"{key}_{count}_duplicate"

    if new_key in dictionary:
        return unique_key(key, dictionary, count + 1)

    return new_key



from __future__ import annotations

from datetime import datetime, timezone, tzinfo
from typing import TYPE_CHECKING
from zoneinfo import ZoneInfo

from dissect.target.plugin import Plugin, internal

if TYPE_CHECKING:
    from dissect.target.target import Target


class DateTimePlugin(Plugin):
    """Generic implementation for datetime plugin."""

    __namespace__ = "datetime"

    def __init__(self, target: Target):
        super().__init__(target)

        if self.__class__ == DateTimePlugin:
            self.target.log.warning(
                "No OS specific datetime information available, falling back to UTC for datetime helpers"
            )

    def check_compatible(self) -> None:
        pass

    @internal
    def tz(self, name: str) -> tzinfo:
        return ZoneInfo(name)

    @internal(property=True)
    def tzinfo(self) -> tzinfo:
        """Return a ``datetime.tzinfo`` of the current system timezone."""
        return timezone.utc

    @internal
    def local(self, dt: datetime) -> datetime:
        """Replace the ``tzinfo`` of a given ``datetime.datetime`` object with the current system ``tzinfo``.

        Does not perform any conversion.
        """
        return dt.replace(tzinfo=self.tzinfo)

    @internal
    def to_utc(self, dt: datetime) -> datetime:
        """Convert any ``datetime.datetime`` object into a UTC ``datetime.datetime`` object.

        First replaces the current ``tzinfo`` with the system ``tzinfo`` without conversion, then converts it to an
        aware UTC ``datetime`` object.
        """
        return self.local(dt).astimezone(timezone.utc)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

import pytest

from dissect.target.filesystems.config import (
    ConfigurationEntry,
    ConfigurationFilesystem,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.fixture
def etc_directory(tmp_path: Path, fs_unix: VirtualFilesystem) -> VirtualFilesystem:
    etc_path = tmp_path.joinpath("etc/")
    etc_path.joinpath("new/path").mkdir(parents=True, exist_ok=True)
    etc_path.joinpath("new/config").mkdir(parents=True, exist_ok=True)
    etc_path.joinpath("new/path/config").write_text(Path(absolute_path("_data/helpers/configutil/config")).read_text())
    fs_unix.map_dir("/etc", etc_path)

    return fs_unix


@pytest.fixture
def mapped_file(test_file: str, fs_unix: VirtualFilesystem) -> VirtualFilesystem:
    local_path = Path(absolute_path(test_file))
    file_name = f"/etc/{local_path.name}"
    fs_unix.map_file(file_name, local_path.absolute())
    return "/" + local_path.name


@pytest.mark.parametrize(
    ("test_file", "expected_output"),
    [
        (
            "_data/helpers/configutil/hosts",
            {
                "127.0.0.1": "localhost",
                "::1": "localhost",
                "127.0.1.1": "pop-os.localdomain pop-os",
            },
        ),
        (
            "_data/helpers/configutil/hosts.allow",
            {"ALL": ["LOCAL @some_netgroup", ".foobar.edu EXCEPT terminalserver.foobar.edu"]},
        ),
        (
            "_data/helpers/configutil/hosts.deny",
            {"ALL": "PARANOID", "ALL EXCEPT in.fingerd": "other.host.name, .other.domain"},
        ),
        (
            "_data/helpers/configutil/resolv.conf",
            {"nameserver": "127.0.0.53", "options": "edns0", "search": "local"},
        ),
        (
            "_data/helpers/configutil/sshd_config",
            {
                "HostKey": [f"__PROGRAMDATA__/ssh/ssh_host_{key}_key" for key in ["rsa", "dsa", "ecdsa", "ed25519"]],
                "AuthorizedKeysFile": ".ssh/authorized_keys",
                "Subsystem": "sftp sftp-server.exe",
                "Match Group administrators": {
                    "AuthorizedKeysFile": "__PROGRAMDATA__/ssh/administrators_authorized_keys"
                },
                "Match User anoncvs": {
                    "AllowTcpForwarding": "no",
                    "PermitTTY": "no",
                    "ForceCommand": "cvs server",
                },
            },
        ),
        (
            "_data/helpers/configutil/test.xml",
            {
                "Server": {
                    "attributes": {"port": "8005", "shutdown": "SHUTDOWN"},
                    "nodes": {
                        "Listener": {
                            "attributes": {"className": "org.apache.catalina.core.JasperListener1"},
                            "text": "a",
                        },
                        "Listener-2": {
                            "attributes": {"className": "org.apache.catalina.core.JasperListener2"},
                            "text": "b",
                        },
                        "Service": {
                            "attributes": {"name": "Catalina"},
                            "nodes": {
                                "Connector": {
                                    "attributes": {
                                        "port": "8080",
                                        "protocol": "HTTP/1.1",
                                        "connectionTimeout": "20000",
                                        "redirectPort": "8443",
                                    },
                                },
                                "Engine": {
                                    "attributes": {"name": "Catalina", "defaultHost": "localhost"},
                                    "nodes": {
                                        "Host": {
                                            "attributes": {
                                                "name": "localhost",
                                                "appBase": "webapps",
                                                "unpackWARs": "true",
                                                "autoDeploy": "true",
                                            },
                                            "nodes": {
                                                "Valve": {
                                                    "attributes": {
                                                        "className": "org.apache.catalina.valves.AccessLogValve",
                                                        "directory": "logs",
                                                        "prefix": "localhost_access_log.",
                                                        "suffix": ".txt",
                                                        "pattern": "%h %l %u %t s",
                                                    },
                                                }
                                            },
                                        }
                                    },
                                },
                            },
                        },
                    },
                },
            },
        ),
    ],
)
def test_parse_file_input(target_unix: Target, mapped_file: str, expected_output: dict) -> None:
    config_filesystem = ConfigurationFilesystem(target_unix, "/etc")
    entry: ConfigurationEntry = config_filesystem.get(mapped_file)
    check_dictionary(expected_output, entry.parser_items)


def check_dictionary(expected_dict: dict, data_dict: dict) -> None:
    for key, value in expected_dict.items():
        if info_value := data_dict.get(key):
            check_value(value, info_value)
        else:
            raise AssertionError(f"Key {key!r} was not found in parser_items.")


def check_value(expected_value: Any, value: Any) -> None:
    assert type(expected_value) is type(value), "The types of the values are not the same"
    if isinstance(expected_value, list):
        # Check if all elements of the expected value are in value
        assert all(val in value for val in expected_value)

    if isinstance(expected_value, dict):
        check_dictionary(expected_value, value)


def test_unix_registry(target_unix: Target, etc_directory: VirtualFilesystem) -> None:
    config_fs = ConfigurationFilesystem(target_unix, "/etc")
    config_path = list(config_fs.get("/").iterdir())

    assert config_path == ["new"]
    assert sorted(config_fs.get("/new").iterdir()) == ["config", "path"]
    assert isinstance(config_fs.get("/new/path/config"), ConfigurationEntry)


def test_parse_functions(target_unix: Target, etc_directory: VirtualFilesystem) -> None:
    config_fs = ConfigurationFilesystem(target_unix, "/etc")
    entry: ConfigurationEntry = config_fs.get("/new/path/config", collapse=True, separator=(r"\s",))

    assert entry["help"] == "you"
    assert entry["test"] == "you"

    entry = config_fs.get("/new/path/config", collapse={"help"}, separator=(r"\s",))

    assert entry["help"] == "you"
    assert entry["test"] == ["me", "you"]



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.filesystems.ntfs import NtfsFilesystem
from dissect.target.plugins.filesystem.ntfs.usnjrnl import UsnjrnlPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_usnjrnl_normal(target_win: Target) -> None:
    """Test parsing of a usnjrnl file."""
    filesystem = NtfsFilesystem(usnjrnl=absolute_path("_data/plugins/filesystem/ntfs/usnjrnl/usnjrnl.bin").open("rb"))
    target_win.filesystems = [filesystem]

    target_win.add_plugin(UsnjrnlPlugin)

    data = list(target_win.usnjrnl())

    assert len(data) == 15214



from __future__ import annotations

import os
from typing import TYPE_CHECKING, BinaryIO

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path


class DirectoryFilesystem(Filesystem):
    __type__ = "dir"

    def __init__(self, path: Path, *args, **kwargs):
        super().__init__(None, *args, **kwargs)
        self.base_path = path

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.base_path}>"

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        raise TypeError("Detect is not allowed on DirectoryFilesystem class")

    def _resolve_path(self, path: str) -> Path:
        if not self.case_sensitive:
            searchpath = self.base_path

            for p in path.split("/"):
                match = [d for d in searchpath.iterdir() if d.name.lower() == p.lower()]

                if not match or len(match) > 1:
                    raise FileNotFoundError(path)

                searchpath = match[0]

            entry = searchpath
        else:
            entry = self.base_path.joinpath(path.strip("/"))

        return entry

    def get(self, path: str) -> FilesystemEntry:
        if not (path := path.strip("/")):
            return DirectoryFilesystemEntry(self, "/", self.base_path)

        entry = self._resolve_path(path)

        try:
            entry.lstat()
            return DirectoryFilesystemEntry(self, path, entry)
        except Exception:
            raise FileNotFoundError(path)


class DirectoryFilesystemEntry(FilesystemEntry):
    entry: Path

    def get(self, path: str) -> FilesystemEntry:
        path = fsutil.join(self.path, path, alt_separator=self.fs.alt_separator)
        return self.fs.get(path)

    def open(self) -> BinaryIO:
        try:
            if self.is_dir():
                raise IsADirectoryError(self.path)
            return self._resolve().entry.open("rb")
        except (PermissionError, OSError) as e:
            raise FilesystemError from e

    def iterdir(self) -> Iterator[str]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        if self.is_symlink():
            yield from self.readlink_ext().iterdir()
        else:
            for item in self.entry.iterdir():
                yield item.name

    def scandir(self) -> Iterator[FilesystemEntry]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        if self.is_symlink():
            yield from self.readlink_ext().scandir()
        else:
            for item in self.entry.iterdir():
                path = fsutil.join(self.path, item.name, alt_separator=self.fs.alt_separator)
                yield DirectoryFilesystemEntry(self.fs, path, item)

    def exists(self) -> bool:
        try:
            return self._resolve().entry.exists()
        except FilesystemError:
            return False

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.is_dir()
        except FilesystemError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.is_file()
        except FilesystemError:
            return False

    def is_symlink(self) -> bool:
        try:
            return self.entry.is_symlink()
        except (FilesystemError, OSError):
            return False

    def readlink(self) -> str:
        if not self.is_symlink():
            raise NotASymlinkError

        # We want to get the "truest" form of the symlink
        # If we use the readlink() of pathlib.Path directly, it gets thrown into the path parsing of pathlib
        # Because DirectoryFilesystem may also be used with TargetPath, we specifically handle that case here
        # and use os.readlink for host paths
        if isinstance(self.entry, fsutil.TargetPath):
            return self.entry.get().readlink()
        return os.readlink(self.entry)  # noqa: PTH115

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self._resolve(follow_symlinks=follow_symlinks).entry.lstat()

    def lstat(self) -> fsutil.stat_result:
        return fsutil.stat_result.copy(self.entry.lstat())

    def attr(self) -> dict[str, bytes]:
        return fsutil.fs_attrs(self.entry, follow_symlinks=True)

    def lattr(self) -> dict[str, bytes]:
        return fsutil.fs_attrs(self.entry, follow_symlinks=False)



from __future__ import annotations

import io
import re
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.containers.raw import RawContainer
from dissect.target.tools.qfind import main as target_qfind

if TYPE_CHECKING:
    from dissect.target.target import Target

re_ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")

buf = (
    (b"\x00" * 1024 * 64)
    + (b"ABCD" + b"\x00" * ((1024 * 4) - 4))
    + (b"A\x00B\x00C\x00D\x00" + b"\x00" * ((1024 * 4) - 8))
    + (b"\xab\xcd" + b"\x00" * ((1024 * 4) - 2))
    + (b"\x00" * 1024 * 52)
)

buf_unique = (b"\x00" * 512) + (b"ABCD" + b"\x00" * ((1024 * 8) - 4)) + (b"ABCD" + b"\x00" * ((1024 * 8) - 4))


@pytest.mark.parametrize(
    ("buf", "argv", "expected_out", "not_expected_out", "expected_count"),
    [
        pytest.param(
            buf,
            ["target-qfind", "example.img", "--needles", "ABCD"],
            [
                "[0x010000 @ ABCD (utf-8)]\nABCD\n",
                "[0x012000 @ ABCD (hex)]\n\n",  # No printable string
            ],
            None,
            None,
            id="input-simple",
        ),
        pytest.param(
            buf,
            ["target-qfind", "example.img", "--needles", "B"],
            [
                "[0x010001 @ B (utf-8)]",
                "ABCD",
            ],
            None,
            None,
            id="input-short",
        ),
        pytest.param(
            buf,
            ["target-qfind", "example.img", "--needles", "ABCD", "--raw"],
            [
                "[0x010000 @ ABCD (utf-8)]",
                "00010000  41 42 43 44 00 00 00 00  00 00 00 00 00 00 00 00   ABCD............",
                "[0x012000 @ ABCD (hex)]",
                "00012000  ab cd 00 00 00 00 00 00  00 00 00 00 00 00 00 00   ................",
            ],
            None,
            None,
            id="output-raw",
        ),
        pytest.param(
            buf,
            ["target-qfind", "example.img", "--needles", "ABCD", "--encoding", "utf-16-le", "--raw"],
            [
                "[0x010000 @ ABCD (utf-8)]",
                "00010000  41 42 43 44 00 00 00 00  00 00 00 00 00 00 00 00   ABCD............",
                "[0x011000 @ ABCD (utf-16-le)]",
                "00011000  41 00 42 00 43 00 44 00  00 00 00 00 00 00 00 00   A.B.C.D.........",
                "[0x012000 @ ABCD (hex)]",
                "00012000  ab cd 00 00 00 00 00 00  00 00 00 00 00 00 00 00   ................",
            ],
            None,
            None,
            id="input-utf-16-le",
        ),
        pytest.param(
            buf,
            ["target-qfind", "example.img", "--needles", "ABCD", "--no-hex-decode", "--raw"],
            [
                "[0x010000 @ ABCD (utf-8)]",
            ],
            [
                "[0x012000 @ ABCD (hex)]",
            ],
            None,
            id="behavior-no-hex-decode",
        ),
        pytest.param(
            buf,
            ["target-qfind", "example.img", "--needles", "abcd", "--ignore-case", "--raw"],
            [
                "[0x010000 @ abcd (utf-8)]",
                "00010000  41 42 43 44 00 00 00 00  00 00 00 00 00 00 00 00   ABCD............",
            ],
            None,
            None,
            id="behavior-ignore-case-output-raw",
        ),
        pytest.param(
            buf,
            ["target-qfind", "example.img", "--needles", r"[a-d]{4}", "--regex", "--ignore-case", "--raw"],
            [
                "[0x010000 @ [a-d]{4} (utf-8)]",
                "00010000  41 42 43 44 00 00 00 00  00 00 00 00 00 00 00 00   ABCD............",
            ],
            None,
            None,
            id="input-regex-behavior-ignore-case-output-raw",
        ),
        pytest.param(
            buf_unique,
            ["target-qfind", "example.img", "--needles", "ABCD"],
            None,
            None,
            [("ABCD", 4)],  # two headers and two recovered strings
            id="output-not-unique",
        ),
        pytest.param(
            buf_unique,
            ["target-qfind", "example.img", "--needles", "ABCD", "--unique"],
            None,
            None,
            [("ABCD", 2)],
            id="output-unique",  # one header and one recovered string
        ),
    ],
)
def test_qfind(
    target_bare: Target,
    buf: bytes,
    argv: list[str],
    expected_out: list[str],
    not_expected_out: list[str] | None,
    expected_count: list[tuple] | None,
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture,
) -> None:
    """Test if target-qfind behaves as expected by inspecting stdout contents."""

    mock_disk = RawContainer(io.BytesIO(buf))
    target_bare.disks.add(mock_disk)

    with patch("dissect.target.Target.open_all", return_value=[target_bare]), monkeypatch.context() as m:
        m.setenv("NO_COLOR", "1")
        m.setattr("sys.argv", argv)
        target_qfind()

        out = re_ansi_escape.sub("", capsys.readouterr().out)

        if expected_out:
            for exp_out in expected_out:
                assert exp_out in out

        if not_expected_out:
            for ne_out in not_expected_out:
                assert ne_out not in out

        if expected_count:
            for c_out, count in expected_count:
                assert out.count(c_out) == count



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.os.windows.dpapi.dpapi import DPAPIPlugin
from dissect.target.plugins.os.windows.dpapi.keyprovider.empty import EmptyKeyProviderPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


def test_dpapi_keyprovider_empty(target_win: Target) -> None:
    """test if we yield an empty key correctly."""

    target_win.add_plugin(DPAPIPlugin, check_compatible=False)
    target_win.add_plugin(EmptyKeyProviderPlugin)

    key = next(target_win.dpapi.keyprovider.empty())

    assert key == ("dpapi.keyprovider.empty", "")



from __future__ import annotations

import io
from typing import TYPE_CHECKING
from urllib.parse import urlparse

import pytest
from flow.record.fieldtypes import path as flow_path

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.loaders.log import LogLoader

if TYPE_CHECKING:
    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("path", "uri", "input_file", "expected_mapping"),
    [
        ("/dir/*.evt*", None, "/dir/test.evtx", "/sysvol/windows/system32/winevt/logs/test.evtx"),
        ("/dir/*.evt*", None, "/dir/test.evt", "/sysvol/windows/system32/config/test.evt"),
        ("/source/iis.log", "log:///dir/with/files/*.log?hint=iis", "/source/iis.log", "/sysvol/files/logs/iis.log"),
    ],
)
def test_log_loader(target_default: Target, path: str, uri: str, input_file: str, expected_mapping: str) -> None:
    with pytest.deprecated_call(match="The LogLoader is deprecated in favor of direct files"):
        vfs = VirtualFilesystem()
        vfs.map_file_fh(input_file, io.BytesIO(b"\x00"))
        log_loader = LogLoader(vfs.path(path), parsed_path=urlparse(uri))
        log_loader.map(target_default)
        # TODO: RGLOB does not take into account the seperator of the target, so maybe an issue with the flavour?
        observed_mapping = next(target_default.fs.path("/").rglob("*.*"))
        assert expected_mapping == flow_path(observed_mapping)



from __future__ import annotations

import logging
import socket
import ssl
import time
import warnings
from io import DEFAULT_BUFFER_SIZE
from struct import pack, unpack
from typing import TYPE_CHECKING

from dissect.util.stream import AlignedStream

from dissect.target.containers.raw import RawContainer
from dissect.target.exceptions import LoaderError
from dissect.target.loader import Loader
from dissect.target.plugin import arg

if TYPE_CHECKING:
    import urllib.parse
    from pathlib import Path

    from dissect.target.target import Target

log = logging.getLogger(__name__)


class RemoteStream(AlignedStream):
    def __init__(self, stream: RemoteStreamConnection, disk_id: int, size: int | None = None):
        self.stream = stream
        self.disk_id = disk_id
        super().__init__(size)

    def _read(self, offset: int, length: int) -> bytes:
        return self.stream.read(self.disk_id, offset, length)


class RemoteStreamConnection:
    # Max. number of times we try to reconnect (still tweaking this)
    MAX_RECONNECTS = 30

    # Max. number of short reads, short reads might happen because of internal bugs
    # this is a mechanism to make sure the client will not just hang
    MAX_SHORT_READS = 10

    # Time to wait before attempting to reconnect
    RECONNECT_WAIT = 10

    # Socket timeout, connections can be slow, so not too low
    # Also there might be an initial waiting time of 10s because of the
    # previous connection! So it must be at least 10s.
    SOCKET_TIMEOUT = 30

    # Remote agent understands 3 commands:
    # 1      INFO: return disk size and sector size for each remote disk
    # 2      QUIT: stops the agent on the remote machine
    # 50 + X READ: read disk number X (starts with 0)
    COMMAND_INFO = 1
    COMMAND_QUIT = 2
    COMMAND_READ = 50

    # This values can be injected by other tooling to let the Remote Loader
    # take advantage of embedded SSL artefacts.
    CONFIG_KEY = None
    CONFIG_CRT = None

    @staticmethod
    def configure(key: str, crt: str) -> None:
        RemoteStreamConnection.CONFIG_KEY = key
        RemoteStreamConnection.CONFIG_CRT = crt

    def __init__(self, hostname: str, port: int, **kwargs):
        self.hostname = hostname
        self.port = port
        self._is_connected = False
        self._socket = None
        self._ssl_sock = None
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            self._context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        self._context.verify_mode = ssl.CERT_REQUIRED
        self._context.load_default_certs()
        self._max_reconnects = self.MAX_RECONNECTS
        self._max_shortreads = self.MAX_SHORT_READS
        self._reconnect_wait = self.RECONNECT_WAIT
        self._socket_timeout = self.SOCKET_TIMEOUT

        flag_cert_chain_loaded = False
        flag_verify_locations_loaded = False

        if options := kwargs.get("options"):
            client_key = options.get("key")
            client_crt = options.get("crt")
            server_ca = options.get("ca")
            noverify = options.get("noverify")

            if client_key and client_crt:
                self._context.load_cert_chain(certfile=client_crt, keyfile=client_key)
                flag_cert_chain_loaded = True
            if noverify:
                self._context.verify_mode = ssl.CERT_NONE
            if server_ca:
                self._context.load_verify_locations(server_ca)
                flag_verify_locations_loaded = True
            self._max_reconnects = options.get("reconnects", max(0, self._max_reconnects))
            self._max_shortreads = options.get("shortreads", max(0, self._max_shortreads))
            self._reconnect_wait = options.get("reconnectwait", max(0, self._reconnect_wait))
            self._socket_timeout = options.get("sockettimeout", max(0, self._socket_timeout))

        if (
            flag_cert_chain_loaded is False
            and self.CONFIG_KEY is not None
            and self.CONFIG_CRT is not None
            and hasattr(self._context, "load_cert_chain_str")
        ):
            self._context.load_cert_chain_str(certfile=self.CONFIG_CRT, keyfile=self.CONFIG_KEY)

        if flag_verify_locations_loaded is False and self.CONFIG_CRT is not None:
            self._context.load_verify_locations(cadata=self.CONFIG_CRT)

        self.log = log

    def is_connected(self) -> bool:
        return self._is_connected

    def connect(self) -> None:
        if self._is_connected:
            return

        reconnects = 0
        while self._is_connected is False:
            self.log.debug("Connecting to agent")
            # Even during the handshake things can go wrong with unreliable connections
            try:
                self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self._socket.settimeout(self._socket_timeout)
                ssl_sock = self._context.wrap_socket(self._socket, server_hostname=self.hostname)
                ssl_sock.connect((self.hostname, self.port))
                self._ssl_sock = ssl_sock
                self._is_connected = True
                self.log.debug("Connection established with agent")
            except Exception:
                # If the max. connections are exceeded it is probably no use anymore because
                # the remote system has become unreachable and we need to report back to the
                # user that we can no longer contact the remote machine.
                if reconnects > self._max_reconnects:
                    raise ConnectionError("Unable to reconnect with remote agent.")
                if self._ssl_sock is not None:
                    self._ssl_sock.close()
                if self._socket is not None:
                    self._socket.close()
                # Directly re-connecting seem to be less succesful, allow some time to re-connect
                # seems to yield best results in practice
                self.log.debug("Unable to connect to agent, next attempt in %d sec", self._reconnect_wait)
                time.sleep(self._reconnect_wait)
                reconnects += 1

    def _receive_bytes(self, length: int) -> bytes:
        data = b""
        received = 0
        short_reads = 0
        while received < length:
            packet = self._ssl_sock.recv(min(length - received, DEFAULT_BUFFER_SIZE))
            packet_size = len(packet)
            data += packet
            received += packet_size
            if packet_size == 0:
                short_reads += 1
            if short_reads > self._max_shortreads:
                raise TimeoutError("Too many short reads.")

        return data

    def read(self, disk_id: int, offset: int, length: int) -> bytes:
        # In practice this should not occur but if one uses this stream directly we will
        # be kind and inform the user about the limitations of this logic
        if length < 0:
            raise NotImplementedError("RemoteStreamConnection does not support length = -1")

        if length == 0:
            return b""

        data = b""
        received = 0
        while received < length:
            self.connect()
            self._ssl_sock.send(pack(">BQQ", self.COMMAND_READ + disk_id, offset, length))
            try:
                data += self._receive_bytes(length)
                received += length
            except Exception:
                self.log.debug("Unable to read data from agent, re-connecting")
                self._ssl_sock.close()
                self._socket.close()
                self._is_connected = False

        return data

    def close(self) -> None:
        if self.is_connected:
            self._ssl_sock.send(pack(">BQQ", self.COMMAND_QUIT, 0, 0))

    def info(self) -> list[RemoteStream]:
        self.connect()
        response_size = 16
        self._ssl_sock.send(pack(">BQQ", self.COMMAND_INFO, 0, 0))
        number_of_disks = unpack("<B", self._ssl_sock.recv(1))[0]
        remainder = number_of_disks * response_size
        data = self._receive_bytes(remainder)
        disks = []
        for i in range(0, number_of_disks * response_size, response_size):
            part = data[i : i + response_size]
            (disk_size, _) = unpack("<QQ", part)
            disks.append(RemoteStream(self, i // response_size, disk_size))

        return disks


@arg("--remote-key", dest="key", help="private key")
@arg("--remote-crt", dest="crt", help="client certificate")
@arg("--remote-ca", dest="ca", help="certificate Authority")
@arg("--remote-noverify", dest="noverify", help="no certificate verification")
@arg("--remote-reconnects", dest="reconnects", help="max number of reconnects")
@arg("--remote-shortreads", dest="shortreads", help="max limit shortreads")
@arg("--remote-reconnectwait", dest="reconnectwait", help="max time before reconnection attempt")
@arg("--remote-sockettimeout", dest="sockettimeout", help="socket timeout")
class RemoteLoader(Loader):
    """Load a remote target that runs a compatible Dissect agent."""

    def __init__(self, path: Path, parsed_path: urllib.parse.ParseResult | None = None):
        super().__init__(path, parsed_path=parsed_path, resolve=False)
        if parsed_path is None:
            raise LoaderError("No URI connection details has been passed.")
        self.stream = RemoteStreamConnection(parsed_path.hostname, parsed_path.port, options=self.parsed_query)

    @staticmethod
    def detect(path: Path) -> bool:
        # You can only activate this loader by URI-scheme "remote://"
        return False

    def map(self, target: Target) -> None:
        self.stream.log = target.log
        for disk in self.stream.info():
            target.disks.add(RawContainer(disk))



from __future__ import annotations

import os
import time
from functools import cached_property
from typing import TYPE_CHECKING

from dissect.regf import regf
from dissect.util import ts

from dissect.target.exceptions import (
    LoaderError,
    RegistryKeyNotFoundError,
    RegistryValueNotFoundError,
)
from dissect.target.filesystems.smb import SmbFilesystem
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.regutil import (
    RegistryHive,
    RegistryKey,
    RegistryValue,
    ValueType,
)
from dissect.target.loader import Loader
from dissect.target.plugins.os.windows._os import WindowsPlugin
from dissect.target.plugins.os.windows.registry import RegistryPlugin

if TYPE_CHECKING:
    from datetime import datetime
    from pathlib import Path
    from urllib.parse import ParseResult

    from impacket.dcerpc.v5.srvs import SHARE_INFO_1

    from dissect.target.target import Target

try:
    from impacket.dcerpc.v5 import rpcrt, rrp, scmr, transport
    from impacket.dcerpc.v5.rpcrt import DCERPCException
    from impacket.smbconnection import SessionError, SMBConnection

    HAS_IMPACKET = True
except ImportError:
    HAS_IMPACKET = False


class SmbLoader(Loader):
    """Use remote SMB servers as targets.

    This loader maps SMB shares from a remote SMB server as filesystems. It makes
    use of Impacket's ``SMBConnection`` class to connect to the remote server and
    enumerate all available shares. The shares are then mapped as filesystems
    in the target. You can use the following examples to connect to a guest.

    Connect as guest (you're probably not going to get very far with this one)::

        smb://10.10.128.3

    Connect as administrator using NTLM password authentication::

        smb://administrator:Dissect123!@10.10.128.3

    Connect as administrator using NTLM pass-the-hash authentication::

        smb://administrator@10.10.128.3?hash=3c19c73ccd2bbcb84c592321caa4b1be

    Connect using a Kerberos authentication::

        smb://administrator@infected.dissect.lab?kerberos=true&ip=10.10.128.3&ticket=administrator.ccache&dc=dc01.dissect.lab&dc-ip=10.10.10.10

    You can also use the following environment variables instead:

    - ``SMB_TARGET_IP``
    - ``SMB_TARGET_HOST``
    - ``SMB_DOMAIN``
    - ``SMB_USERNAME``
    - ``SMB_PASSWORD``
    - ``SMB_KERBEROS_TICKET (or KRB5CCNAME)``
    - ``SMB_KERBEROS_DC``
    - ``SMB_KERBEROS_DC_IP``
    - ``SMB_KERBEROS_AES_KEY``

    When using environment variables, you can invoke this loader using an empty ``smb://`` URI.
    """

    MACHINE_NAME = "DISSECT-TARGET"
    EMPTY_NT = "31d6cfe0d16ae931b73c59d7e0c089c0"
    EMPTY_LM = "aad3b435b51404eeaad3b435b51404ee"

    def __init__(self, path: Path, parsed_path: ParseResult | None = None):
        super().__init__(path, parsed_path=parsed_path, resolve=False)

        if not HAS_IMPACKET:
            raise ImportError(
                "Required dependency 'impacket' is missing, install with 'pip install dissect.target[smb]'"
            )

        if parsed_path is None:
            raise LoaderError("Missing URI connection details")

        self._ip = self.parsed_query.get("ip", os.getenv("SMB_TARGET_IP", parsed_path.hostname))
        self._host = self.parsed_query.get("host", os.getenv("SMB_TARGET_HOST", parsed_path.hostname))
        self._domain = self.parsed_query.get("domain", os.getenv("SMB_DOMAIN", "."))
        self._username = parsed_path.username or os.getenv("SMB_USERNAME", "Guest")
        self._password = parsed_path.password or os.getenv("SMB_PASSWORD", "")
        self._nt, self._lm = "", ""
        if not self._password:
            self._nt, self._lm = self._get_hashes()

        krb_ticketparsed_query = self.parsed_query.get("ticket", self.parsed_query.get("ccache", ""))
        krb_ticket_env = os.getenv("SMB_KERBEROS_TICKET", os.getenv("KRB5CCNAME", ""))
        self._krb_ticket = krb_ticketparsed_query or krb_ticket_env
        self._krb_aes = self.parsed_query.get("aes", os.getenv("SMB_KERBEROS_AES_KEY", ""))
        self._krb_dc = self.parsed_query.get("dc", os.getenv("SMB_KERBEROS_DC", ""))
        self._krb_dc_ip = self.parsed_query.get("dc-ip", os.getenv("SMB_KERBEROS_DC_IP", ""))
        self._use_kerberos = self.parsed_query.get("kerberos", "").lower() in ("true", "1", "yes")

        self._conn = SMBConnection(
            remoteName=self._host,
            remoteHost=self._ip,
            myName=self.MACHINE_NAME,
        )

        if self._use_kerberos:
            # Hack to make Impacket load our kerberos ticket
            os.environ["KRB5CCNAME"] = self._krb_ticket

            # Perform kerberos login
            self._conn.kerberosLogin(
                domain=self._domain,
                user=self._username,
                password=self._password,
                nthash=self._nt,
                lmhash=self._lm,
                aesKey=self._krb_aes,
                kdcHost=self._krb_dc_ip,
                useCache=bool(self._krb_ticket),
            )
        else:
            self._conn.login(
                domain=self._domain,
                user=self._username,
                password=self._password,
                nthash=self._nt,
                lmhash=self._lm,
            )

    def _get_hashes(self) -> tuple[str, str]:
        """Attempt to parse NT and LM hashes from the URI query string."""
        nt = self.parsed_query.get("nt", self.EMPTY_NT)
        lm = self.parsed_query.get("lm", self.EMPTY_LM)

        hashes = self.parsed_query.get("hash", self.parsed_query.get("hashes", f"{nt}:{lm}"))

        if ":" in hashes:
            nt, lm = hashes.split(":", 1)
        else:
            nt = hashes

        return nt, lm

    @staticmethod
    def detect(path: Path) -> bool:
        """This loader can only be activated with the URI-scheme ``smb://<ip>``."""
        return False

    def map(self, target: Target) -> None:
        """Map all target filesystems (network shares) from the SMB connection."""
        target.log.debug("Attempting to list shares...")
        shares: list[SHARE_INFO_1] = self._conn.listShares()
        for share in shares:
            share_name: str = share["shi1_netname"][:-1]
            try:
                smb_filesystem = SmbFilesystem(self._conn, share_name)
                target.filesystems.add(smb_filesystem)

                mount_name = share_name
                if len(share_name) == 2:
                    mount_name = share_name.lower().replace("$", ":")

                target.fs.mount(mount_name, smb_filesystem)
            except SessionError as e:
                target.log.warning("Failed to mount share '%s', reason: %s", share_name, e)

        target.add_plugin(SmbRegistry(target, self._conn), check_compatible=False)
        target._os_plugin = WindowsPlugin


class SmbRegistry(RegistryPlugin):
    __register__ = False

    def __init__(self, target: Target, conn: SMBConnection):
        self.conn = conn
        self._svc_handle = None
        self._was_stopped = False
        self._was_disabled = False
        self._svcctl = None
        self._winreg = None
        super().__init__(target)

    def check_compatible(self) -> bool:
        return False

    def _init_registry(self) -> None:
        try:
            self._svcctl = _connect_rpc(self.conn, "ncacn_np:445[\\pipe\\svcctl]", scmr.MSRPC_UUID_SCMR)
            self._check_service_status()

            self._winreg = _connect_rpc(self.conn, "ncacn_np:445[\\pipe\\winreg]", rrp.MSRPC_UUID_RRP)

            hklm_hive = SmbRegistryHive(
                self._winreg, "HKEY_LOCAL_MACHINE", rrp.hOpenLocalMachine(self._winreg)["phKey"]
            )
            hku_hive = SmbRegistryHive(self._winreg, "HKEY_USERS", rrp.hOpenUsers(self._winreg)["phKey"])

            self._add_hive("HKLM", hklm_hive, TargetPath(self.target.fs, "HKLM"))
            self._add_hive("HKU", hku_hive, TargetPath(self.target.fs, "HKU"))
            self._map_hive("HKEY_LOCAL_MACHINE", hklm_hive)
            self._map_hive("HKEY_USERS", hku_hive)
        except SessionError:
            self.target.log.info("Failed to open remote registry, registry will not be available")
            return  # no registry access, probably no access rights

    def _init_users(self) -> None:
        pass

    def __del__(self) -> None:
        if hasattr(self, "_was_stopped") and self._was_stopped:
            scmr.hRControlService(self._svcctl, self._svc_handle, scmr.SERVICE_CONTROL_STOP)

        if hasattr(self, "_was_disabled") and self._was_disabled:
            scmr.hRChangeServiceConfigW(self._svcctl, self._svc_handle, dwStartType=0x4)

        if getattr(self, "_svcctl", None):
            self._svcctl.disconnect()

        if getattr(self, "_winreg", None):
            self._winreg.disconnect()

    def _check_service_status(self) -> None:
        try:
            manager_handle = scmr.hROpenSCManagerW(self._svcctl)["lpScHandle"]
            self._svc_handle = scmr.hROpenServiceW(self._svcctl, manager_handle, "RemoteRegistry")["lpServiceHandle"]
        except DCERPCException:
            return

        current_state = scmr.hRQueryServiceStatus(self._svcctl, self._svc_handle)["lpServiceStatus"]["dwCurrentState"]
        if current_state == scmr.SERVICE_STOPPED:
            self._was_stopped = True

            start_type = scmr.hRQueryServiceConfigW(self._svcctl, self._svc_handle)["lpServiceConfig"]["dwStartType"]
            if start_type == 0x4:
                self._was_disabled = True
                scmr.hRChangeServiceConfigW(self._svcctl, self._svc_handle, dwStartType=0x3)

            scmr.hRStartServiceW(self._svcctl, self._svc_handle)
            time.sleep(1)


class SmbRegistryHive(RegistryHive):
    def __init__(self, winreg: rpcrt.DCERPC_v5, name: str, handle: int):
        self.winreg = winreg
        self.name = name
        self.handle = handle

    def key(self, key: str) -> SmbRegistryKey:
        return SmbRegistryKey(self, key)


class SmbRegistryKey(RegistryKey):
    hive: SmbRegistryHive

    def __init__(self, hive: SmbRegistryHive, path: str):
        self._path = path
        self._name: str = path.rsplit("\\", 1)[-1]
        super().__init__(hive)

    @cached_property
    def handle(self) -> int:
        try:
            return rrp.hBaseRegOpenKey(
                self.hive.winreg,
                self.hive.handle,
                self._path,
                samDesired=rrp.MAXIMUM_ALLOWED | rrp.KEY_ENUMERATE_SUB_KEYS | rrp.KEY_QUERY_VALUE,
            )["phkResult"]
        except rrp.DCERPCSessionError:
            raise RegistryKeyNotFoundError(self.path)

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> str:
        return f"{self.hive.name}\\{self._path}" if self._path else self.hive.name

    @property
    def timestamp(self) -> datetime:
        return ts.from_unix(0)

    def subkey(self, subkey: str) -> SmbRegistryKey:
        # To improve peformance, immediately return a "hollow" key object
        # Only listing all subkeys or reading a value will result in data being loaded
        # Technically this means we won't raise a RegistryKeyNotFoundError in the correct place
        return SmbRegistryKey(self.hive, f"{self._path}\\{subkey}" if self._path else subkey)

    def subkeys(self) -> list[SmbRegistryKey]:
        subkeys = []
        handle = self.handle

        i = 0
        while True:
            try:
                name = rrp.hBaseRegEnumKey(self.hive.winreg, handle, i)["lpNameOut"][:-1]
                subkeys.append(self.subkey(name))
                i += 1
            except Exception:  # noqa: PERF203
                break

        return subkeys

    def _value(self, value: str) -> SmbRegistryValue:
        reg_value = value.lower()
        for val in self.values():
            if val.name.lower() == reg_value:
                return val
        else:
            raise RegistryValueNotFoundError(value)

    def values(self) -> list[SmbRegistryValue]:
        values = []
        handle = self.handle

        i = 0
        while True:
            try:
                result = rrp.hBaseRegEnumValue(self.hive.winreg, handle, i)
                values.append(
                    SmbRegistryValue(
                        self.hive,
                        result["lpValueNameOut"][:-1] or "(Default)",
                        b"".join(result["lpData"]),
                        result["lpType"],
                    )
                )
                i += 1
            except Exception:  # noqa: PERF203
                break

        return values


class SmbRegistryValue(RegistryValue):
    def __init__(self, hive: str, name: str, data: bytes, type: int):
        super().__init__(hive)
        self._name = name
        self._type = type
        self._value = regf.parse_value(type, data)

    @property
    def name(self) -> str:
        return self._name

    @property
    def value(self) -> ValueType:
        return self._value

    @property
    def type(self) -> str:
        return self._type


def _connect_rpc(conn: SMBConnection, binding: str, uuid: bytes) -> rpcrt.DCERPC_v5:
    rpc = transport.DCERPCTransportFactory(binding)
    rpc.set_smb_connection(conn)
    dce = rpc.get_dce_rpc()
    dce.connect()
    dce.bind(uuid)
    return dce



from __future__ import annotations

import base64
import logging
from typing import TYPE_CHECKING

from defusedxml import ElementTree
from dissect.hypervisor import hyperv

from dissect.target import container
from dissect.target.helpers import fsutil
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from pathlib import Path
    from xml.etree.ElementTree import Element

    from dissect.target.target import Target

log = logging.getLogger(__name__)

DRIVE_CONTROLLER_GUIDS = [
    # Microsoft Emulated IDE Controller
    "83f8638b-8dca-4152-9eda-2ca8b33039b4",
    # Synthetic SCSI Controller
    "d422512d-2bf2-4752-809d-7b82b5fcb1b4",
]


def is_hyperv_xml(path: Path) -> bool:
    """Return if an XML file is a valid Hyper-V configuration.

    Args:
        path: Path to the file to check.
    """
    if path.suffix != ".xml":
        return False

    try:
        et = ElementTree.fromstring(path.read_bytes())
    except Exception:
        return False
    else:
        element_names = [el.tag for el in et]
        return et.tag == "configuration" and "manifest" in element_names


def xml_as_dict(element: Element, root: dict | None = None) -> dict:
    """Convert a Hyper-V XML file into a dictionary.

    Recursively converts all XML elements into a correctly typed dictionary.

    Args:
        element: The current element to convert.
        root: The dictionary object to use as current root.
    """
    root = {} if root is None else root

    ftype = element.get("type")
    if ftype is None:
        obj = root.setdefault(element.tag, {})
        for el in element:
            xml_as_dict(el, obj)
    elif ftype == "string":
        root[element.tag] = element.text
    elif ftype == "integer":
        root[element.tag] = int(element.text)
    elif ftype == "bytes":
        root[element.tag] = base64.b64decode(element.text)
    elif ftype in ("bool", "boolean"):
        root[element.tag] = element.text.lower() == "true"
    else:
        # We don't necessarily want to error out, so add as string instead
        log.warning("Unknown Hyper-V XML value type, adding as string instead: %s (%s)", ftype, element.tag)
        root[element.tag] = element.text

    return root


class HyperVLoader(Loader):
    """Load Microsoft Hyper-V hypervisor files.

    Maps all virtual disks to the target. All paths are absolute in Hyper-V configuration files, so we first attempt
    to locate a file with the same name in the same path as the configuration file. This is the most common method if
    we get a copy of a Hyper-V VM. If that fails, we fall back to the absolute path, which is required when we're
    targetting a Hyper-V host and loading child VMs.

    References:
        - https://learn.microsoft.com/en-us/virtualization/hyper-v-on-windows/about/
    """

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)

        self.vmcx = None
        self.xml = None

        if path.suffix == ".vmcx":
            self.vmcx = hyperv.HyperVFile(path.open("rb"))
        elif path.suffix == ".xml":
            self.xml = ElementTree.fromstring(path.read_bytes())
        else:
            raise ValueError(f"HyperVLoader initialized with unsupported file: {path}")

    @staticmethod
    def detect(path: Path) -> bool:
        return path.suffix.lower() == ".vmcx" or is_hyperv_xml(path)

    def map(self, target: Target) -> None:
        obj = self.vmcx.as_dict() if self.vmcx else xml_as_dict(self.xml)

        configuration = obj["configuration"]

        # Naive approach first
        instance_guids = {key for key in configuration if key[1:-1].lower() in DRIVE_CONTROLLER_GUIDS}

        # Add properly discovered instances next
        for value in configuration["manifest"].values():
            # Iterate all vdevXXX entries
            if not isinstance(value, dict) or "device" not in value:
                continue

            if value["device"].lower() in DRIVE_CONTROLLER_GUIDS:
                instance_guids.add(f"_{value['instance']}_")

        if not instance_guids:
            raise ValueError("Unable to find drive controllers")

        for guid in instance_guids:
            # The GUID can be stored in lower or upper case
            device = configuration.get(guid, configuration.get(guid.lower(), {}))

            for key, value in device.items():
                if not key.startswith("controller"):
                    continue

                for drive in value.values():
                    if drive["type"] != "VHD":
                        continue

                    if not drive["pathname"]:
                        continue

                    filepath = drive["pathname"]
                    filename = fsutil.basename(filepath, alt_separator="\\")

                    # First attempt path relative to .vmcx/.xml file
                    disk_path = self.base_path.joinpath(filename)

                    # Next attempt "Virtual Hard Disks" directory relative from .vmcx/.xml file
                    if not disk_path.exists():
                        disk_path = self.base_path.joinpath("Virtual Hard Disks").joinpath(filename)

                    # Next attempt "Virtual Hard Disks" directory one up from .vmcx/.xml file
                    if not disk_path.exists():
                        disk_path = self.base_path.parent.joinpath("Virtual Hard Disks").joinpath(filename)

                    # Finally attempt absolute path
                    if not disk_path.exists():
                        disk_path = self.base_path.joinpath("/").joinpath(filepath).resolve()

                    try:
                        target.disks.add(container.open(disk_path))
                    except Exception:
                        # In some situations we may have something like a .vhdx extension but a raw disk inside
                        # Fall back on stream based container detection
                        try:
                            target.disks.add(container.open(disk_path.open("rb")))
                        except Exception:
                            target.log.exception("Failed to load VHD: %s", drive)



from __future__ import annotations

import inspect
import itertools
import textwrap
from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from dissect.target.plugin import Plugin

NO_DOCS = "No documentation"

FUNCTION_OUTPUT_DESCRIPTION = {
    "none": "no output",
    "default": "text",
    "record": "records",
    "yield": "lines",
}

INDENT_STEP = " " * 4

FUNC_DOC_TEMPLATE = "{func_name} - {short_description} (output: {output_type})"


def get_docstring(obj: Any, placeholder: str = NO_DOCS) -> str:
    """Get object's docstring or a placeholder if no docstring found."""
    # Use of `inspect.cleandoc()` is preferred to `textwrap.dedent()` here
    # because many multi-line docstrings in the codebase
    # have no indentation in the first line, which confuses `dedent()`
    return inspect.cleandoc(obj.__doc__) if obj.__doc__ else placeholder


def get_func_description(func: Callable, with_docstrings: bool = False) -> str:
    klass, func = _get_real_func_obj(func)
    func_output, func_doc = _get_func_details(func)

    # get user-friendly function name
    func_name = _get_full_func_name(klass, func)

    if with_docstrings:
        func_title = f"`{func_name}` (output: {func_output})"
        func_doc = textwrap.indent(func_doc, prefix=INDENT_STEP)
        desc = f"{func_title}\n\n{func_doc}"
    else:
        docstring_first_line = func_doc.splitlines()[0].lstrip()
        desc = FUNC_DOC_TEMPLATE.format(
            func_name=func_name, short_description=docstring_first_line, output_type=func_output
        )

    return desc


def get_plugin_functions_desc(plugin_class: type[Plugin], with_docstrings: bool = False) -> str:
    descriptions = []
    for func_name in plugin_class.__exports__:
        func_obj = getattr(plugin_class, func_name)
        if func_obj is getattr(plugin_class, "__call__", None):  # noqa: B004
            continue

        _, func = _get_real_func_obj(func_obj)
        func_desc = get_func_description(func, with_docstrings=with_docstrings)
        descriptions.append(func_desc)

    # sort functions in the plugin alphabetically
    descriptions = sorted(descriptions)

    if with_docstrings:
        # add empty lines after every func description
        descriptions = [block for pair in zip(descriptions, itertools.repeat("")) for block in pair]

    return "\n".join(descriptions)


def get_plugin_description(plugin_class: type[Plugin]) -> str:
    plugin_name = plugin_class.__name__
    plugin_desc_title = f"`{plugin_name}` (`{plugin_class.__module__}.{plugin_name}`)"
    plugin_doc = textwrap.indent(get_docstring(plugin_class), prefix=INDENT_STEP)
    return f"{plugin_desc_title}\n\n{plugin_doc}"


def get_plugin_overview(
    plugin_class: type[Plugin], with_plugin_desc: bool = False, with_func_docstrings: bool = False
) -> str:
    paragraphs = []

    if with_plugin_desc:
        plugin_desc = get_plugin_description(plugin_class)
        paragraphs.append(plugin_desc)

        # Add a line break between plugin description and function descriptions
        paragraphs.append("")

    func_descriptions_paragraph = get_plugin_functions_desc(plugin_class, with_docstrings=with_func_docstrings)

    # Indent func docs additionally if plugin desc is printed
    if with_plugin_desc:
        func_descriptions_paragraph = textwrap.indent(
            func_descriptions_paragraph,
            prefix=INDENT_STEP,
        )

    paragraphs.append(func_descriptions_paragraph)
    return "\n".join(paragraphs)


def _get_plugin_class_for_func(func: Callable) -> type[Plugin]:
    """Return plugin class for provided function instance."""
    func_parent_name = func.__qualname__.rsplit(".", 1)[0]
    return getattr(inspect.getmodule(func), func_parent_name, None)


def _get_real_func_obj(func: Callable) -> tuple[type[Plugin], Callable]:
    """Return a tuple with plugin class and underlying function object for provided function instance."""
    klass = None

    if isinstance(func, property):
        # turn property into function
        func = func.fget

    if inspect.ismethod(func):
        for klass in func.__self__.__class__.__mro__:
            if func.__name__ in klass.__dict__:
                break
        else:
            func = getattr(func, "__func__", func)

    if inspect.isfunction(func):
        klass = _get_plugin_class_for_func(func)

    if not klass:
        raise ValueError(f"Can't find class for {func}")

    return (klass, func)


def _get_func_details(func: Callable) -> tuple[str, str]:
    """Return a tuple with function's name, output label and docstring."""
    func_doc = get_docstring(func)

    if hasattr(func, "__output__") and func.__output__ in FUNCTION_OUTPUT_DESCRIPTION:
        func_output = FUNCTION_OUTPUT_DESCRIPTION[func.__output__]
    else:
        func_output = "unknown"

    return (func_output, func_doc)


def _get_full_func_name(plugin_class: type[Plugin], func: Callable) -> str:
    func_name = func.__name__

    if hasattr(plugin_class, "__namespace__") and plugin_class.__namespace__:
        func_name = f"{plugin_class.__namespace__}.{func_name}"

    return func_name



from __future__ import annotations

import datetime
import logging
import re
from dataclasses import dataclass
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self

    from dissect.target.target import Target

log = logging.getLogger(__name__)

SchedLgURecord = TargetRecordDescriptor(
    "windows/tasks/log/schedlgu",
    [
        ("datetime", "ts"),
        ("string", "job"),
        ("string", "command"),
        ("string", "status"),
        ("uint32", "exit_code"),
        ("string", "version"),
    ],
)

JOB_REGEX_PATTERN = re.compile(r"\"(.*?)\" \((.*?)\)")
SCHEDLGU_REGEX_PATTERN = re.compile(r"\".+\n.+\n\s{4}.+\n|\".+\n.+", re.MULTILINE)


@dataclass(order=True)
class SchedLgU:
    ts: datetime.datetime | None = None
    job: str | None = None
    status: str | None = None
    command: str | None = None
    exit_code: int | None = None
    version: str | None = None

    @staticmethod
    def _sanitize_ts(ts: str, tzinfo: datetime.tzinfo = datetime.timezone.utc) -> datetime:
        # sometimes "at" exists before the timestamp
        ts = ts.strip("at ")
        try:
            ts = datetime.datetime.strptime(ts, "%m/%d/%Y %I:%M:%S %p").replace(tzinfo=tzinfo)
        except ValueError:
            ts = datetime.datetime.strptime(ts, "%d-%m-%Y %H:%M:%S").replace(tzinfo=tzinfo)

        return ts

    @staticmethod
    def _parse_job(line: str) -> tuple[str, str | None]:
        matches = JOB_REGEX_PATTERN.match(line)
        if matches:
            return matches.groups()

        log.warning("SchedLgU failed to parse job and command from line: %r, returning line", line)
        return line, None

    @classmethod
    def from_line(cls, line: str, tzinfo: datetime.tzinfo = datetime.timezone.utc) -> Self:
        """Parse a group of SchedLgU.txt lines."""
        event = cls()
        lines = line.splitlines()

        # Events can have 2 or 3 lines as a group in total. An example of a complete task job event is:
        # "Symantec NetDetect.job" (NDETECT.EXE)
        #     Finished 14-9-2003 13:21:01
        #     Result: The task completed with an exit code of (65).
        if len(lines) == 3:
            event.job, event.command = cls._parse_job(lines[0])
            event.status, event.ts = lines[1].split(maxsplit=1)
            event.exit_code = int(lines[2].split("(")[1].rstrip(")."))

        # Events that have 2 lines as a group can be started task job event or the Task Scheduler Service. Examples:
        #   "Symantec NetDetect.job" (NDETECT.EXE)
        #        Started at 14-9-2003 13:26:00
        elif len(lines) == 2 and ".job" in lines[0]:
            event.job, event.command = cls._parse_job(lines[0])
            event.status, event.ts = lines[1].split(maxsplit=1)

        # Events without a task job event are the Task Scheduler Service events. Which can look like this:
        # "Task Scheduler Service"
        #      Exited at 14-9-2003 13:40:24
        # OR
        # "Task Scheduler Service"
        # 6.0.6000.16386 (vista_rtm.061101-2205)
        elif len(lines) == 2:
            event.job = lines[0].strip('"')

            if lines[1].startswith("\t") or lines[1].startswith(" "):
                event.status, event.ts = lines[1].split(maxsplit=1)
            else:
                event.version = lines[1]

        if event.ts:
            event.ts = cls._sanitize_ts(event.ts, tzinfo=tzinfo)

        return event


class SchedLgUPlugin(Plugin):
    """Plugin for parsing the Task Scheduler Service transaction log file (SchedLgU.txt)."""

    PATHS = (
        "sysvol/SchedLgU.txt",
        "%windir%/SchedLgU.txt",
        "%windir%/tasks/SchedLgU.txt",
    )

    def __init__(self, target: Target):
        self.target = target
        self.paths = [resolved_path for path in self.PATHS if (resolved_path := self.target.resolve(path)).exists()]

    def check_compatible(self) -> None:
        if len(self.paths) == 0:
            raise UnsupportedPluginError("No SchedLgU.txt file found.")

    @export(record=SchedLgURecord)
    def schedlgu(self) -> Iterator[SchedLgURecord]:
        """Return all events in the Task Scheduler Service transaction log file (SchedLgU.txt).

        Older Windows systems may log ``.job`` tasks that get started remotely in the SchedLgU.txt file.
        In addition, this log file records when the Task Scheduler service starts and stops.

        Adversaries may use malicious ``.job`` files to gain persistence on a system.

        Yields SchedLgURecord with fields:

        .. code-block:: text

            ts (datetime): The timestamp of the event.
            job (str): The name of the .job file.
            command (str): The command executed.
            status (str): The status of the event (finished, completed, exited, stopped).
            exit_code (int): The exit code of the event.
            version (str): The version of the Task Scheduler service.
        """
        target_tz = self.target.datetime.tzinfo

        for path in self.paths:
            content = path.read_text(encoding="UTF-16", errors="surrogateescape")

            for match in re.findall(SCHEDLGU_REGEX_PATTERN, content):
                event = SchedLgU.from_line(match, tzinfo=target_tz)

                yield SchedLgURecord(
                    ts=event.ts,
                    job=event.job,
                    command=event.command,
                    status=event.status,
                    exit_code=event.exit_code,
                    version=event.version,
                    _target=self.target,
                )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers import configutil
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, alias, export
from dissect.target.plugins.general.users import UserDetails

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target

TrashRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "linux/filesystem/recyclebin",
    [
        ("datetime", "ts"),
        ("path", "path"),
        ("filesize", "filesize"),
        ("path", "deleted_path"),
        ("path", "source"),
    ],
)


class GnomeTrashPlugin(Plugin):
    """Linux GNOME Trash plugin."""

    PATHS = (
        # Default $XDG_DATA_HOME/Trash
        ".local/share/Trash",
    )

    def __init__(self, target: Target):
        super().__init__(target)
        self.trashes = set(self._garbage_collector())

    def _garbage_collector(self) -> Iterator[tuple[UserDetails, TargetPath]]:
        """it aint much, but its honest work"""

        # home trash folders
        for user_details in self.target.user_details.all_with_home():
            for trash_path in self.PATHS:
                if (path := user_details.home_path.joinpath(trash_path)).exists():
                    yield user_details, path

        # mounted devices trash folders
        for mount_path in [*self.target.fs.mounts, "/mnt", "/media"]:
            if mount_path == "/":
                continue

            for mount_trash in self.target.fs.path(mount_path).rglob(".Trash-*"):
                yield UserDetails(None, None), mount_trash

    def check_compatible(self) -> None:
        if not self.trashes:
            raise UnsupportedPluginError("No Trash folder(s) found")

    @export(record=TrashRecord)
    @alias(name="recyclebin")
    def trash(self) -> Iterator[TrashRecord]:
        """Yield deleted files from GNOME Trash folders.

        Recovers deleted files and artifacts from ``$HOME/.local/share/Trash``.
        Probably also works with other desktop interfaces as long as they follow the Trash specification from FreeDesktop.

        Also parses media trash locations such as ``/media/$USER/$Label/.Trash-*``, ``/mnt/$Label/.Trash-*`` and other
        locations as defined in ``/etc/fstab``.

        References:
            - https://specifications.freedesktop.org/trash-spec/latest/
            - https://github.com/GNOME/glib/blob/main/gio/glocalfile.c
            - https://specifications.freedesktop.org/basedir-spec/latest/

        Yields ``TrashRecord`` records with the following fields:

        .. code-block:: text

            ts           (datetime): timestamp when the file was deleted or for expunged files when it could not be permanently deleted
            path         (path):     path where the file was located before it was deleted
            filesize     (filesize): size in bytes of the deleted file
            deleted_path (path):     path to the current location of the deleted file
            source       (path):     path to the .trashinfo file
        """  # noqa: E501

        for user_details, trash in self.trashes:
            for trash_info_file in trash.glob("info/*.trashinfo"):
                trash_info = configutil.parse(trash_info_file, hint="ini").get("Trash Info", {})
                original_path = self.target.fs.path(trash_info.get("Path", ""))

                # We use the basename of the .trashinfo file and not the Path variable inside the
                # ini file. This way we can keep duplicate basenames of trashed files separated correctly.
                deleted_path = trash / "files" / trash_info_file.name.replace(".trashinfo", "")

                if deleted_path.exists():
                    deleted_files = [deleted_path]

                    if deleted_path.is_dir():
                        deleted_files.extend(deleted_path.rglob("*"))

                    for file in deleted_files:
                        # NOTE: We currently do not 'fix' the original_path of files inside deleted directories.
                        # This would require guessing where the parent folder starts, which is impossible without
                        # making assumptions.
                        yield TrashRecord(
                            ts=trash_info.get("DeletionDate", 0),
                            path=original_path,
                            filesize=file.lstat().st_size if file.is_file() else None,
                            deleted_path=file,
                            source=trash_info_file,
                            _user=user_details.user,
                            _target=self.target,
                        )

                # We cannot determine if the deleted entry is a directory since the path does
                # not exist at $TRASH/files, so we work with what we have instead.
                else:
                    self.target.log.warning("Expected trashed file(s) at %s", deleted_path)
                    yield TrashRecord(
                        ts=trash_info.get("DeletionDate", 0),
                        path=original_path,
                        filesize=0,
                        deleted_path=deleted_path,
                        source=trash_info_file,
                        _user=user_details.user,
                        _target=self.target,
                    )

            # We also iterate expunged folders, they can contain files that could not be
            # deleted when the user pressed the "empty trash" button in the file manager.
            # References:
            #   - https://gitlab.gnome.org/GNOME/glib/-/issues/1665
            #   - https://bugs.launchpad.net/ubuntu/+source/nautilus/+bug/422012
            for item in (trash / "expunged").rglob("*/*"):
                stat = item.lstat()
                yield TrashRecord(
                    ts=stat.st_mtime,  # NOTE: This is the timestamp at which the file failed to delete
                    path=None,  # We do not know the original file path
                    filesize=stat.st_size if item.is_file() else None,
                    deleted_path=item,
                    source=trash / "expunged",
                    _user=user_details.user,
                    _target=self.target,
                )



"""A pathlib.Path compatible implementation for dissect.target.

This allows for the majority of the pathlib.Path API to "just work" on dissect.target filesystems.

Most of this consists of subclassed internal classes with dissect.target specific patches,
but sometimes the change to a function is small, so the entire internal function is copied
and only a small part changed. To ease updating this code, the order of functions, comments
and code style is kept largely the same as the original pathlib.py.

Yes, we know, this is playing with fire and it can break on new CPython releases.

The implementation is split up in multiple files, one for each CPython version.
You're currently looking at the CPython 3.10 implementation.

Commit hash we're in sync with: b382bf5
"""

from __future__ import annotations

import fnmatch
import re
from pathlib import Path, PurePath, _Accessor, _PosixFlavour
from typing import IO, TYPE_CHECKING, Any, Callable, ClassVar

from dissect.target import filesystem
from dissect.target.exceptions import FilesystemError, SymlinkRecursionError
from dissect.target.helpers import polypath
from dissect.target.helpers.compat import path_common

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem, FilesystemEntry
    from dissect.target.helpers.fsutil import stat_result


class _DissectFlavour(_PosixFlavour):
    is_supported = True

    __variant_instances: ClassVar[dict[tuple[bool, str], _DissectFlavour]] = {}

    def __new__(cls, case_sensitive: bool = False, alt_separator: str = ""):
        idx = (case_sensitive, alt_separator)
        instance = cls.__variant_instances.get(idx, None)
        if instance is None:
            instance = _PosixFlavour.__new__(cls)
            cls.__variant_instances[idx] = instance

        return instance

    def __init__(self, case_sensitive: bool = False, alt_separator: str = ""):
        super().__init__()
        self.altsep = alt_separator
        self.case_sensitive = case_sensitive

    def casefold(self, s: str) -> str:
        return s if self.case_sensitive else s.lower()

    def casefold_parts(self, parts: list[str]) -> list[str]:
        return parts if self.case_sensitive else [p.lower() for p in parts]

    def compile_pattern(self, pattern: str) -> Callable[..., Any]:
        return re.compile(fnmatch.translate(pattern), 0 if self.case_sensitive else re.IGNORECASE).fullmatch


class _DissectAccessor(_Accessor):
    @staticmethod
    def stat(path: TargetPath, *, follow_symlinks: bool = True) -> stat_result:
        if follow_symlinks:
            return path.get().stat()
        return path.get().lstat()

    @staticmethod
    def open(
        path: TargetPath,
        mode: str = "rb",
        buffering: int = 0,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> IO:
        """Open file and return a stream.

        Supports a subset of features of the real pathlib.open/io.open.

        Note: in contrast to regular Python, the mode is binary by default. Text mode
        has to be explicitly specified. Buffering is also disabled by default.
        """
        return path_common.io_open(path, mode, buffering, encoding, errors, newline)

    @staticmethod
    def listdir(path: TargetPath) -> list[str]:
        return path.get().listdir()

    @staticmethod
    def scandir(path: TargetPath) -> path_common._DissectScandirIterator:
        return path_common.scandir(path)

    @staticmethod
    def chmod(path: TargetPath, mode: int, *, follow_symlinks: bool = True) -> None:
        raise NotImplementedError("TargetPath.chmod() is unsupported")

    @staticmethod
    def lchmod(path: TargetPath, mode: int) -> None:
        raise NotImplementedError("TargetPath.lchmod() is unsupported")

    @staticmethod
    def mkdir(path: TargetPath, mode: int = 0o777, parents: bool = False, exist_ok: bool = False) -> None:
        raise NotImplementedError("TargetPath.mkdir() is unsupported")

    @staticmethod
    def unlink(path: TargetPath, missing_ok: bool = False) -> None:
        raise NotImplementedError("TargetPath.unlink() is unsupported")

    @staticmethod
    def link(src: str, dst: TargetPath) -> None:
        raise NotImplementedError("TargetPath.link() is unsupported")

    @staticmethod
    def rmdir(path: TargetPath) -> None:
        raise NotImplementedError("TargetPath.rmdir() is unsupported")

    @staticmethod
    def rename(path: TargetPath, target: str) -> str:
        raise NotImplementedError("TargetPath.rename() is unsupported")

    @staticmethod
    def replace(path: TargetPath, target: str) -> str:
        raise NotImplementedError("TargetPath.replace() is unsupported")

    @staticmethod
    def symlink(path: TargetPath, target: str, target_is_directory: bool = False) -> None:
        raise NotImplementedError("TargetPath.symlink() is unsupported")

    @staticmethod
    def touch(path: TargetPath, mode: int = 0o666, exist_ok: bool = True) -> None:
        raise NotImplementedError("TargetPath.touch() is unsupported")

    @staticmethod
    def readlink(path: TargetPath) -> str:
        return path.get().readlink()

    @staticmethod
    def owner(path: TargetPath) -> str:
        raise NotImplementedError("TargetPath.owner() is unsupported")

    @staticmethod
    def group(path: TargetPath) -> str:
        raise NotImplementedError("TargetPath.group() is unsupported")

    @staticmethod
    def getcwd() -> str:
        raise NotImplementedError("TargetPath.getcwd() is unsupported")

    @staticmethod
    def expanduser(path: str) -> str:
        raise NotImplementedError("TargetPath.expanduser() is unsupported")

    realpath = staticmethod(path_common.realpath)

    # NOTE: Forward compatibility with CPython >= 3.12
    isjunction = staticmethod(path_common.isjunction)


_dissect_accessor = _DissectAccessor()


class PureDissectPath(PurePath):
    _fs: Filesystem
    _flavour = _DissectFlavour(case_sensitive=False)

    def __reduce__(self) -> tuple:
        raise TypeError("TargetPath pickling is currently not supported")

    @classmethod
    def _from_parts(cls, args: list) -> Self:
        fs = args[0]

        if not isinstance(fs, filesystem.Filesystem):
            raise TypeError(
                "invalid PureDissectPath initialization: missing filesystem, "
                f"got {args!r} (this might be a bug, please report)"
            )

        alt_separator = fs.alt_separator
        path_args = []
        for arg in args[1:]:
            if isinstance(arg, str):
                arg = polypath.normalize(arg, alt_separator=alt_separator)
            path_args.append(arg)

        self = super()._from_parts(path_args)
        self._fs = fs

        self._flavour = _DissectFlavour(alt_separator=fs.alt_separator, case_sensitive=fs.case_sensitive)

        return self

    def _make_child(self, args: list) -> Self:
        child = super()._make_child(args)
        child._fs = self._fs
        child._flavour = self._flavour
        return child

    def with_name(self, name: str) -> Self:
        result = super().with_name(name)
        result._fs = self._fs
        result._flavour = self._flavour
        return result

    def with_stem(self, stem: str) -> Self:
        result = super().with_stem(stem)
        result._fs = self._fs
        result._flavour = self._flavour
        return result

    def with_suffix(self, suffix: str) -> Self:
        result = super().with_suffix(suffix)
        result._fs = self._fs
        result._flavour = self._flavour
        return result

    def relative_to(self, *other) -> Self:
        result = super().relative_to(*other)
        result._fs = self._fs
        result._flavour = self._flavour
        return result

    def __rtruediv__(self, key: str) -> Self:
        try:
            return self._from_parts([self._fs, key, *self._parts])
        except TypeError:
            return NotImplemented

    @property
    def parent(self) -> Self:
        result = super().parent
        result._fs = self._fs
        result._flavour = self._flavour
        return result

    @property
    def parents(self) -> path_common._DissectPathParents:
        return path_common._DissectPathParents(self)


class TargetPath(Path, PureDissectPath):
    _accessor = _dissect_accessor
    __slots__ = ("_entry",)

    def _make_child_relpath(self, part: str) -> Self:
        child = super()._make_child_relpath(part)
        child._fs = self._fs
        child._flavour = self._flavour
        return child

    def get(self) -> FilesystemEntry:
        try:
            return self._entry
        except AttributeError:
            self._entry = self._fs.get(str(self))
            return self._entry

    @classmethod
    def cwd(cls) -> Self:
        """Return a new path pointing to the current working directory
        (as returned by os.getcwd()).
        """
        raise NotImplementedError("TargetPath.cwd() is unsupported")

    @classmethod
    def home(cls) -> Self:
        """Return a new path pointing to the user's home directory (as
        returned by os.path.expanduser('~')).
        """
        raise NotImplementedError("TargetPath.home() is unsupported")

    def iterdir(self) -> Iterator[Self]:
        """Iterate over the files in this directory.  Does not yield any
        result for the special paths '.' and '..'.
        """
        for entry in self._accessor.scandir(self):
            if entry.name in {".", ".."}:
                # Yielding a path object for these makes little sense
                continue
            child_path = self._make_child_relpath(entry.name)
            child_path._entry = entry
            yield child_path

    # NOTE: Forward compatibility with CPython >= 3.12
    def walk(
        self, top_down: bool = True, on_error: Callable[[Exception], None] | None = None, follow_symlinks: bool = False
    ) -> Iterator[tuple[Self, list[str], list[str]]]:
        """Walk the directory tree from this directory, similar to os.walk()."""
        paths = [self]

        while paths:
            path = paths.pop()
            if isinstance(path, tuple):
                yield path
                continue

            # We may not have read permission for self, in which case we can't
            # get a list of the files the directory contains. os.walk()
            # always suppressed the exception in that instance, rather than
            # blow up for a minor reason when (say) a thousand readable
            # directories are still left to visit. That logic is copied here.
            try:
                scandir_it = self._accessor.scandir(path)
            except OSError as e:
                if on_error is not None:
                    on_error(e)
                continue

            with scandir_it:
                dirnames = []
                filenames = []
                for entry in scandir_it:
                    try:
                        is_dir = entry.is_dir(follow_symlinks=follow_symlinks)
                    except OSError:
                        # Carried over from os.path.isdir().
                        is_dir = False

                    if is_dir:
                        dirnames.append(entry.name)
                    else:
                        filenames.append(entry.name)

            if top_down:
                yield path, dirnames, filenames
            else:
                paths.append((path, dirnames, filenames))

            paths += [path._make_child_relpath(d) for d in reversed(dirnames)]

    def absolute(self) -> Self:
        """Return an absolute version of this path.  This function works
        even if the path doesn't point to anything.

        No normalization is done, i.e. all '.' and '..' will be kept along.
        Use resolve() to get the canonical path to a file.
        """
        raise NotImplementedError("TargetPath.absolute() is unsupported in Dissect")

    # NOTE: We changed some of the error handling here to deal with our own exception types
    def resolve(self, strict: bool = False) -> Self:
        """
        Make the path absolute, resolving all symlinks on the way and also
        normalizing it (for example turning slashes into backslashes under
        Windows).
        """

        s = self._accessor.realpath(self, strict=strict)
        p = self._from_parts((self._fs, s))

        # In non-strict mode, realpath() doesn't raise on symlink loops.
        # Ensure we get an exception by calling stat()
        if not strict:
            try:
                p.stat()
            except FilesystemError as e:
                if isinstance(e, SymlinkRecursionError):
                    raise
        return p

    def stat(self, *, follow_symlinks: bool = True) -> stat_result:
        """
        Return the result of the stat() system call on this path, like
        os.stat() does.
        """
        return self._accessor.stat(self, follow_symlinks=follow_symlinks)

    def open(
        self,
        mode: str = "rb",
        buffering: int = 0,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> IO:
        """Open file and return a stream.

        Supports a subset of features of the real pathlib.open/io.open.

        Note: in contrast to regular Python, the mode is binary by default. Text mode
        has to be explicitly specified. Buffering is also disabled by default.
        """
        return self._accessor.open(self, mode, buffering, encoding, errors, newline)

    def write_bytes(self, data: bytes) -> int:
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        raise NotImplementedError("TargetPath.write_bytes() is unsupported")

    def write_text(
        self, data: str, encoding: str | None = None, errors: str | None = None, newline: str | None = None
    ) -> int:
        """
        Open the file in text mode, write to it, and close the file.
        """
        raise NotImplementedError("TargetPath.write_text() is unsupported")

    def readlink(self) -> Self:
        """
        Return the path to which the symbolic link points.
        """
        path = self._accessor.readlink(self)
        return self._from_parts((self._fs, path))

    # NOTE: Forward compatibility with CPython >= 3.12
    def is_junction(self) -> bool:
        """
        Whether this path is a junction.
        """
        return self._accessor.isjunction(self)

    def expanduser(self) -> Self:
        """Return a new path with expanded ~ and ~user constructs
        (as returned by os.path.expanduser)
        """
        raise NotImplementedError("TargetPath.expanduser() is unsupported")



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from defusedxml import ElementTree

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

# Example startupinfo entry:
#
#    <Process Name="C:\Windows\System32\SecurityHealthSystray.exe" PID="6208" StartedInTraceSec="48.500">
#    	<StartTime>2020/09/11:18:12:48.6685573</StartTime>
#    	<CommandLine><![CDATA["C:\Windows\System32\SecurityHealthSystray.exe" ]]></CommandLine>
#    	<DiskUsage Units="bytes">325120</DiskUsage>
#    	<CpuUsage Units="us">32024</CpuUsage>
#    	<ParentPID>6016</ParentPID>
#    	<ParentStartTime>2020/09/11:18:12:30.2666535</ParentStartTime>
#    	<ParentName>explorer.exe</ParentName>
#    </Process>

StartupInfoRecord = TargetRecordDescriptor(
    "filesystem/windows/startupinfo",
    [
        ("datetime", "ts"),
        ("path", "path"),
        ("path", "commandline"),
        ("varint", "pid"),
        ("varint", "parent_pid"),
        ("datetime", "parent_start_time"),
        ("path", "parent_name"),
        ("varint", "disk_usage"),
        ("varint", "cpu_usage"),
    ],
)


def parse_ts(time_string: str) -> datetime.datetime | None:
    if not time_string:
        return None

    return datetime.datetime.strptime(time_string[:26], "%Y/%m/%d:%H:%M:%S.%f").replace(tzinfo=datetime.timezone.utc)


class StartupInfoPlugin(Plugin):
    """Windows startup info plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self._files = []

        path = target.resolve("%windir%/system32/wdi/logfiles/startupinfo")
        if path.exists():
            self._files = list(path.iterdir())

    def check_compatible(self) -> None:
        if not self._files:
            raise UnsupportedPluginError("No StartupInfo files found")

    @export(record=StartupInfoRecord)
    def startupinfo(self) -> Iterator[StartupInfoRecord]:
        """Return the contents of StartupInfo files.

        On a Windows system, the StartupInfo log files contain information about process execution for the first 90
        seconds of user logon activity, such as process name and CPU usage.

        References:
            - https://www.trustedsec.com/blog/who-left-the-backdoor-open-using-startupinfo-for-the-win/
        """
        for path in self._files:
            fh = path.open("rb")

            try:
                root = ElementTree.fromstring(fh.read().decode("utf-16-le"), forbid_dtd=True)
                for process in root.iter("Process"):
                    start_time = process.findtext("StartTime")
                    parent_start_time = process.findtext("ParentStartTime")

                    yield StartupInfoRecord(
                        ts=parse_ts(start_time),
                        path=self.target.fs.path(process.get("Name")),
                        commandline=self.target.fs.path(process.findtext("CommandLine")),
                        pid=process.get("PID"),
                        parent_pid=process.findtext("ParentPID"),
                        parent_start_time=parse_ts(parent_start_time),
                        parent_name=self.target.fs.path(process.findtext("ParentName")),
                        disk_usage=process.findtext("DiskUsage"),
                        cpu_usage=process.findtext("CpuUsage"),
                        _target=self.target,
                    )
            except Exception:
                self.target.log.exception("Failed to open StartupInfo file: %s", str(path))



from __future__ import annotations

import io
import logging
from struct import error as StructError
from typing import TYPE_CHECKING, BinaryIO

from dissect.cstruct import cstruct
from dissect.ole import OLE
from dissect.ole.exceptions import Error as OleError
from dissect.shellitem.lnk import Lnk

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.helpers.shell_application_ids import APPLICATION_IDENTIFIERS
from dissect.target.helpers.utils import findall
from dissect.target.plugin import Plugin, export
from dissect.target.plugins.os.windows.lnk import LnkRecord, parse_lnk_file

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

log = logging.getLogger(__name__)

LNK_GUID = b"\x01\x14\x02\x00\x00\x00\x00\x00\xc0\x00\x00\x00\x00\x00\x00\x46"

JumpListRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/jumplist",
    [
        ("string", "type"),
        ("string", "application_id"),
        ("string", "application_name"),
        *LnkRecord.target_fields,
    ],
)


custom_destination_def = """
struct header {
    int version;
    int unknown1;
    int unknown2;
    int value_type;
}

struct header_end {
    int number_of_entries;
}

struct header_end_0 {
    uint16  name_length;
    wchar   name[name_length];
    int     number_of_entries;
}

struct footer {
    char magic[4];
}
"""

c_custom_destination = cstruct()
c_custom_destination.load(custom_destination_def)


class JumpListFile:
    def __init__(self, fh: BinaryIO, file_name: str):
        self.fh = fh
        self.file_name = file_name

        self.application_id, self.application_type = file_name.split(".")
        self.application_type = self.application_type.split("-")[0]
        self.application_name = APPLICATION_IDENTIFIERS.get(self.application_id)

    def __iter__(self) -> Iterator[Lnk]:
        raise NotImplementedError

    @property
    def name(self) -> str:
        """Return the name of the application."""
        return self.application_name

    @property
    def id(self) -> str:
        """Return the application identifier."""
        return self.application_id

    @property
    def type(self) -> str:
        """Return the type of the Jump List file."""
        return self.application_type


class AutomaticDestinationFile(JumpListFile):
    """Parse Jump List AutomaticDestination file."""

    def __init__(self, fh: BinaryIO, file_name: str):
        super().__init__(fh, file_name)
        self.ole = OLE(self.fh)

    def __iter__(self) -> Iterator[Lnk]:
        for dir_name in self.ole.root.listdir():
            if dir_name == "DestList":
                continue

            dir = self.ole.get(dir_name)

            for item in dir.open():
                try:
                    yield Lnk(io.BytesIO(item))
                except StructError:  # noqa: PERF203
                    continue
                except Exception as e:
                    log.warning("Failed to parse LNK file from directory %s", dir_name)
                    log.debug("", exc_info=e)
                    continue


class CustomDestinationFile(JumpListFile):
    """Parse Jump List CustomDestination file."""

    MAGIC_FOOTER = 0xBABFFBAB
    VERSIONS = (2,)

    def __init__(self, fh: BinaryIO, file_name: str):
        super().__init__(fh, file_name)

        self.fh.seek(-4, io.SEEK_END)
        self.footer = c_custom_destination.footer(self.fh.read(4))
        self.magic = int.from_bytes(self.footer.magic, "little")

        self.fh.seek(0, io.SEEK_SET)
        self.header = c_custom_destination.header(self.fh)
        self.version = self.header.version

        if self.header.value_type == 0:
            self.header_end = c_custom_destination.header_end_0(self.fh)
        elif self.header.value_type in [1, 2]:
            self.header_end = c_custom_destination.header_end(self.fh)
        else:
            raise NotImplementedError(
                f"The value_type ({self.header.value_type}) of the CustomDestination file is not implemented"
            )

        if self.version not in self.VERSIONS:
            raise NotImplementedError(f"The CustomDestination file has an unsupported version: {self.version}")

        if self.magic != self.MAGIC_FOOTER:
            raise ValueError(f"The CustomDestination file has an invalid magic footer: {self.magic}")

    def __iter__(self) -> Iterator[Lnk]:
        # Searches for all LNK GUID's because the number of entries in the header is not always correct.
        buf = self.fh.read()

        for offset in findall(buf, LNK_GUID):
            try:
                lnk = Lnk(io.BytesIO(buf[offset + len(LNK_GUID) :]))
                yield lnk
            except EOFError:  # noqa: PERF203
                break
            except Exception as e:
                log.warning("Failed to parse LNK file from a CustomDestination file")
                log.debug("", exc_info=e)
                continue


class JumpListPlugin(Plugin):
    """Jump List is a Windows feature introduced in Windows 7.

    It stores information about recently accessed applications and files.

    References:
        - https://forensics.wiki/jump_lists
        - https://github.com/libyal/dtformats/blob/main/documentation/Jump%20lists%20format.asciidoc
    """

    __namespace__ = "jumplist"

    def __init__(self, target: Target):
        super().__init__(target)
        self.automatic_destinations = []
        self.custom_destinations = []

        for user_details in self.target.user_details.all_with_home():
            for destination in user_details.home_path.glob(
                "AppData/Roaming/Microsoft/Windows/Recent/CustomDestinations/*.customDestinations-ms"
            ):
                self.custom_destinations.append([destination, user_details.user])

        for user_details in self.target.user_details.all_with_home():
            for destination in user_details.home_path.glob(
                "AppData/Roaming/Microsoft/Windows/Recent/AutomaticDestinations/*.automaticDestinations-ms"
            ):
                self.automatic_destinations.append([destination, user_details.user])

    def check_compatible(self) -> None:
        if not any([self.automatic_destinations, self.custom_destinations]):
            raise UnsupportedPluginError("No Jump List files found")

    @export(record=JumpListRecord)
    def custom_destination(self) -> Iterator[JumpListRecord]:
        """Return the content of CustomDestination Windows Jump Lists.

        These are created when a user pins an application or a file in a Jump List.

        Yields JumpListRecord with fields:

        .. code-block:: text

            type (string): Type of Jump List.
            application_id (string): ID of the application.
            application_name (string): Name of the application.
            lnk_path (path): Path of the link (.lnk) file.
            lnk_name (string): Name of the link (.lnk) file.
            lnk_mtime (datetime): Modification time of the link (.lnk) file.
            lnk_atime (datetime): Access time of the link (.lnk) file.
            lnk_ctime (datetime): Creation time of the link (.lnk) file.
            lnk_relativepath (path): Relative path of target file to the link (.lnk) file.
            lnk_workdir (path): Path of the working directory the link (.lnk) file will execute from.
            lnk_iconlocation (path): Path of the display icon used for the link (.lnk) file.
            lnk_arguments (string): Command-line arguments passed to the target (linked) file.
            local_base_path (string): Absolute path of the target (linked) file.
            common_path_suffix (string): Suffix of the local_base_path.
            lnk_full_path (string): Full path of the linked file. Made from local_base_path and common_path_suffix.
            lnk_net_name (string): Specifies a server share path; for example, "\\\\server\\share".
            lnk_device_name (string): Specifies a device; for example, the drive letter "D:"
            machine_id (string): The NetBIOS name of the machine where the linked file was last known to reside.
            target_mtime (datetime): Modification time of the target (linked) file.
            target_atime (datetime): Access time of the target (linked) file.
            target_ctime (datetime): Creation time of the target (linked) file.
        """
        yield from self._generate_records(self.custom_destinations, CustomDestinationFile)

    @export(record=JumpListRecord)
    def automatic_destination(self) -> Iterator[JumpListRecord]:
        """Return the content of AutomaticDestination Windows Jump Lists.

        These are created automatically when a user opens an application or file.

        Yields JumpListRecord with fields:

        .. code-block:: text

            type (string): Type of Jump List.
            application_id (string): ID of the application.
            application_name (string): Name of the application.
            lnk_path (path): Path of the link (.lnk) file.
            lnk_name (string): Name of the link (.lnk) file.
            lnk_mtime (datetime): Modification time of the link (.lnk) file.
            lnk_atime (datetime): Access time of the link (.lnk) file.
            lnk_ctime (datetime): Creation time of the link (.lnk) file.
            lnk_relativepath (path): Relative path of target file to the link (.lnk) file.
            lnk_workdir (path): Path of the working directory the link (.lnk) file will execute from.
            lnk_iconlocation (path): Path of the display icon used for the link (.lnk) file.
            lnk_arguments (string): Command-line arguments passed to the target (linked) file.
            local_base_path (string): Absolute path of the target (linked) file.
            common_path_suffix (string): Suffix of the local_base_path.
            lnk_full_path (string): Full path of the linked file. Made from local_base_path and common_path_suffix.
            lnk_net_name (string): Specifies a server share path; for example, "\\\\server\\share".
            lnk_device_name (string): Specifies a device; for example, the drive letter "D:"
            machine_id (string): The NetBIOS name of the machine where the linked file was last known to reside.
            target_mtime (datetime): Modification time of the target (linked) file.
            target_atime (datetime): Access time of the target (linked) file.
            target_ctime (datetime): Creation time of the target (linked) file.
        """
        yield from self._generate_records(self.automatic_destinations, AutomaticDestinationFile)

    def _generate_records(
        self,
        destinations: list,
        destination_file: AutomaticDestinationFile | CustomDestinationFile,
    ) -> Iterator[JumpListRecord]:
        for destination, user in destinations:
            fh = destination.open("rb")

            try:
                jumplist = destination_file(fh, destination.name)
            except OleError:
                continue
            except Exception as e:
                self.target.log.warning("Failed to parse Jump List file: %s", destination)
                self.target.log.debug("", exc_info=e)
                continue

            for lnk in jumplist:
                if lnk := parse_lnk_file(self.target, lnk, destination):
                    yield JumpListRecord(
                        type=jumplist.type,
                        application_name=jumplist.name,
                        application_id=jumplist.id,
                        **lnk._asdict(),
                        _user=user,
                        _target=self.target,
                    )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.packagemanager import PackageManagerPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_packagemanager_logs_debian(target_debian: Target, fs_debian: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/os/unix/linux/debian/apt/history.log")
    fs_debian.map_file("/var/log/apt/history.log", data_file)
    target_debian.add_plugin(PackageManagerPlugin)

    results = list(target_debian.packagemanager.logs())
    assert len(results) == 18


def test_packagemanager_logs_redhat(target_redhat: Target, fs_redhat: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/os/unix/linux/redhat/yum/yum.log")
    fs_redhat.map_file("/var/log/yum.log", data_file)
    target_redhat.add_plugin(PackageManagerPlugin)

    results = list(target_redhat.packagemanager.logs())
    assert len(results) == 5


def test_packagemanager_logs_suse(target_suse: Target, fs_suse: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/os/unix/linux/suse/zypp/history")
    fs_suse.map_file("/var/log/zypp/history", data_file)
    target_suse.add_plugin(PackageManagerPlugin)

    results = list(target_suse.packagemanager.logs())
    assert len(results) == 61



from __future__ import annotations

from hashlib import md5, sha256
from struct import pack

try:
    from Crypto.Cipher import AES, ARC4, DES

    HAS_CRYPTO = True
except ImportError:
    HAS_CRYPTO = False

from typing import TYPE_CHECKING

from dissect.cstruct import cstruct
from dissect.util import ts

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

sam_def = """
struct user_F {
  char      unknown1[8];
  uint64    t_last_login;           /* Time of last login */
  char      unknown2[8];
  uint64    t_last_password_set;    /* Time of last password set */
  char      unknown3[8];
  uint64    t_last_incorrect_login; /* Time of last incorrect password */
  int32     rid;
  char      unknown4[4];
  uint16    ACB_bits;               /* Account type and status flags */
  char      unknown5[2];
  uint16    country_code;
  char      unknown6[2];
  uint16    failedcnt;        /* Count of failed logins, if > than policy it is locked. Resets after successful login */
  uint16    logins;           /* Total logins since creation (max. 0xFFFF = 65535) */
  char      unknown7[0xc];
};

#define ACB_DISABLED   0x0001
#define ACB_HOMDIRREQ  0x0002
#define ACB_PWNOTREQ   0x0004
#define ACB_TEMPDUP    0x0008
#define ACB_NORMAL     0x0010
#define ACB_MNS        0x0020
#define ACB_DOMTRUST   0x0040
#define ACB_WSTRUST    0x0080
#define ACB_SVRTRUST   0x0100
#define ACB_PWNOEXP    0x0200
#define ACB_AUTOLOCK   0x0400

// char *acb_fields[16] = {
//    "Disabled" ,
//    "Homedir req." ,
//    "Passwd not req." ,
//    "Temp. duplicate" ,
//    "Normal account" ,
//    "NMS account" ,
//    "Domain trust act." ,
//    "Wks trust act." ,
//    "Srv trust act" ,
//    "Pwd don't expire" ,
//    "Auto lockout" ,
//    "(unknown 0x08)" ,
//    "(unknown 0x10)" ,
//    "(unknown 0x20)" ,
//    "(unknown 0x40)" ,
//    "(unknown 0x80)" ,
// };


struct user_V {

  int unknown1_1;           /* 0x00 - always zero? */
  int unknown1_2;           /* 0x04 - points to username? */
  int unknown1_3;           /* 0x08 - always 0x02 0x00 0x01 0x00 ? */

  int username_ofs;         /* 0x0c */
  int username_len;         /* 0x10 */

  int unknown2_1;           /* 0x14 - always zero? */

  int fullname_ofs;         /* 0x18 */
  int fullname_len;         /* 0x1c */

  int unknown3_1;           /* 0x20 - always zero? */

  int admin_comment_ofs;    /* 0x24 */
  int admin_comment_len;    /* 0x28 */

  int unknown4_1;           /* 0x2c - alway zero? */

  int user_comment_ofs;     /* 0x30 */
  int user_comment_len;     /* 0x34 */

  int unknown5_1;           /* 0x38 - zero? */
  int unknown5_2;           /* 0x3c - to field 8 bytes before hashes */
  int unknown5_3;           /* 0x40 - zero? or size of above? */
  int unknown5_4;           /* 0x44 - zero? */

  int homedir_ofs;          /* 0x48 */
  int homedir_len;          /* 0x4c */

  int unknown6_1;           /* 0x50 - zero? */

  int drvletter_ofs;        /* 0x54 - drive letter for home dir */
  int drvletter_len;        /* 0x58 - len of above, usually 4   */

  int unknown7_1;           /* 0x5c - zero? */

  int logonscr_ofs;         /* 0x60 - users logon script path */
  int logonscr_len;         /* 0x64 - length of string */

  int unknown8_1;           /* 0x68 - zero? */

  int profilep_ofs;         /* 0x6c - profile path string */
  int profilep_len;         /* 0x70 - profile path stringlen */

  int unknown9_1;           /* 0x74 */

  int workstations_ofs;     /* 0x78 */
  int workstations_len;     /* 0x7c */

  int unknowna_1;          /* 0x80 */

  int allowed_hours_ofs;    /* 0x84 */
  int allowed_hours_len;    /* 0x88 */

  int unknownb_1;          /* 0x8c */
  int unknownb_2;          /* 0x90 - pointer to some place before hashes, after comments */
  int unknownb_3;          /* 0x94 - size of above? */
  int unknownb_4;          /* 0x98 - unknown? always 1? */

  int lmpw_ofs;             /* 0x9c */
  int lmpw_len;             /* 0xa0 */

  int unknownc_1;           /* 0xa4 - zero? */

  int ntpw_ofs;             /* 0xa8 */
  int ntpw_len;             /* 0xac */

  int unknownd_1;           /* 0xb0 */
  int unknownd_2;           /* 0xb4 - points to field after hashes */
  int unknownd_3;           /* 0xb8 - size of above field */
  int unknownd_4;           /* 0xbc - zero? */
  int unknownd_5;           /* 0xc0 - points to field after that */
  int unknownd_6;           /* 0xc4 - size of above */
  int unknownd_7;           /* 0xc8 - zero ? */

  char data[4];             /* Data starts here. All pointers above is relative to this,
                               that is V + 0xCC */
};

struct DOMAIN_ACCOUNT_F {
  uint16 revision;                          /* 0x00 */
  uint16 unknown1_1;                        /* 0x02 */
  uint32 unknown1_2;                        /* 0x04 */
  uint64 creation_time;                     /* 0x08 */
  uint64 domain_modified_count;             /* 0x10 */
  uint64 max_password_age;                  /* 0x18 */
  uint64 min_password_age;                  /* 0x20 */
  uint64 force_logoff;                      /* 0x28 */
  uint64 lock_duration;                     /* 0x30 */
  uint64 lock_observation_window;           /* 0x38 */
  uint64 modified_count_at_last_promotion;  /* 0x40 */
  uint32 next_rid;                          /* 0x48 */
  uint32 password_properties;               /* 0x4c */
  uint16 min_password_length;               /* 0x50 */
  uint16 password_history_length;           /* 0x52 */
  uint16 lockout_threshold;                 /* 0x54 */
  uint16 unknown1_3;                        /* 0x56 */
  uint32 server_state;                      /* 0x58 */
  uint16 server_role;                       /* 0x5c */
  uint16 uas_compability_required;          /* 0x5e */
  uint64 unknown2_1;                        /* 0x60 */
  /* char sam_key[];                           0x70, variable size */
};

struct SAM_KEY {      /* size: 64 */
  uint32 revision;    /* 0x00 */
  uint32 length;      /* 0x04 */
  char salt[16];      /* 0x08 */
  char key[16];       /* 0x18 */
  char checksum[16];  /* 0x28 */
  uint64 reserved;    /* 0x38 */
};

struct SAM_KEY_AES {  /* size: >= 32 */
  uint32 revision;     /* 0x00 */
  uint32 length;       /* 0x04 */
  uint32 checksum_len; /* 0x08 */
  uint32 data_len;     /* 0x0c */
  char salt[16];       /* 0x10 */
  /* char data[];         0x20, variable size */
};

struct SAM_HASH {      /* size: 20 */
  uint16 pek_id;       /* 0x00 */
  uint16 revision;     /* 0x02 */
  /* char hash[16];       0x04, variable size */
};

struct SAM_HASH_AES {  /* size: >=24 */
  uint16 pek_id;        /* 0x00 */
  uint16 revision;      /* 0x02 */
  uint32 data_offset;   /* 0x04 */
  char salt[16];        /* 0x08 */
  /* char data[];          0x18, variable size */
};
"""

c_sam = cstruct().load(sam_def)

SamRecord = TargetRecordDescriptor(
    "windows/credential/sam",
    [
        ("datetime", "ts"),
        ("uint32", "rid"),
        ("string", "fullname"),
        ("string", "username"),
        ("string", "admincomment"),
        ("string", "usercomment"),
        ("datetime", "lastlogin"),
        ("datetime", "lastpasswordset"),
        ("datetime", "lastincorrectlogin"),
        ("uint32", "flags"),
        ("uint16", "countrycode"),
        ("uint32", "failedlogins"),
        ("uint32", "logins"),
        ("string", "lm"),
        ("string", "nt"),
    ],
)


def expand_des_key(key: bytes) -> bytes:
    # Expand the key from a 7-byte password key into an 8-byte DES key
    key = bytearray(key[:7]).ljust(7, b"\x00")
    s = bytearray(
        [
            ((key[0] >> 1) & 0x7F) << 1,
            ((key[0] & 0x01) << 6 | ((key[1] >> 2) & 0x3F)) << 1,
            ((key[1] & 0x03) << 5 | ((key[2] >> 3) & 0x1F)) << 1,
            ((key[2] & 0x07) << 4 | ((key[3] >> 4) & 0x0F)) << 1,
            ((key[3] & 0x0F) << 3 | ((key[4] >> 5) & 0x07)) << 1,
            ((key[4] & 0x1F) << 2 | ((key[5] >> 6) & 0x03)) << 1,
            ((key[5] & 0x3F) << 1 | ((key[6] >> 7) & 0x01)) << 1,
            (key[6] & 0x7F) << 1,
        ]
    )
    return bytes(s)


def rid_to_key(rid: int) -> tuple[bytes, bytes]:
    s = rid.to_bytes(4, "little", signed=False)
    k1 = expand_des_key(bytes([s[0], s[1], s[2], s[3], s[0], s[1], s[2]]))
    k2 = expand_des_key(bytes([s[3], s[0], s[1], s[2], s[3], s[0], s[1]]))

    return k1, k2


def decrypt_single_hash(rid: int, samkey: bytes, enc_hash: bytes, apwd: bytes) -> bytes:
    if not enc_hash:
        return b""

    sh = c_sam.SAM_HASH(enc_hash)

    if sh.revision not in [0x01, 0x02]:
        raise ValueError(f"Unsupported LM/NT hash revision encountered: {sh.revision}")

    d1, d2 = (DES.new(k, DES.MODE_ECB) for k in rid_to_key(rid))

    if sh.revision == 0x01:  # LM/NT revision 0x01 involving RC4
        sh_hash = enc_hash[len(c_sam.SAM_HASH) :]
        if not sh_hash:  # Empty hash
            return b""

        rc4_key = md5(samkey + pack("<L", rid) + apwd).digest()
        obfkey = ARC4.new(rc4_key).encrypt(sh_hash)

    else:  # LM/NT revision 0x02 involving AES
        sh = c_sam.SAM_HASH_AES(enc_hash)
        if not sh.data_offset:  # Empty hash
            return b""

        sh_hash = enc_hash[len(c_sam.SAM_HASH_AES) :]
        obfkey = AES.new(samkey, AES.MODE_CBC, sh.salt).decrypt(sh_hash)[:16]

    return d1.decrypt(obfkey[:8]) + d2.decrypt(obfkey[8:])


class SamPlugin(Plugin):
    """SAM plugin.

    References:
        - MS-SAMR Specification
        - Reversing samsrv.dll
        - https://github.com/gentilkiwi/mimikatz
        - https://github.com/skelsec/pypykatz
        - https://web.archive.org/web/20190717124313/http://www.beginningtoseethelight.org/ntsecurity/index.htm
    """

    SAM_KEY = "HKEY_LOCAL_MACHINE\\SAM\\SAM\\Domains\\Account"

    def check_compatible(self) -> None:
        if not HAS_CRYPTO:
            raise UnsupportedPluginError("Missing pycryptodome dependency")

        if not self.target.has_function("lsa"):
            raise UnsupportedPluginError("LSA plugin is required for SAM plugin")

        if not len(list(self.target.registry.keys(self.SAM_KEY))) > 0:
            raise UnsupportedPluginError(f"Registry key not found: {self.SAM_KEY}")

    def calculate_samkey(self, syskey: bytes) -> bytes:
        aqwerty = b"!@#$%^&*()qwertyUIOPAzxcvbnmQQQQQQQQQQQQ)(*@&%\0"
        anum = b"0123456789012345678901234567890123456789\0"

        f_reg = self.target.registry.key(self.SAM_KEY).value("F").value
        f = c_sam.DOMAIN_ACCOUNT_F(f_reg)
        f_key = f_reg[len(c_sam.DOMAIN_ACCOUNT_F) :]
        fk = c_sam.SAM_KEY(f_key)

        if f.revision not in [0x02, 0x03]:
            raise ValueError(f"Unsupported Domain Account F revision encountered: {f.revision}")

        if fk.revision not in [0x01, 0x02]:
            raise ValueError(f"Unsupported SAM Key Data revision encountered: {fk.revision}")

        if fk.revision == 0x01:  # SAM key revision 0x01 involving RC4 (samsrv.dll: KEDecryptKeyWithRC4)
            rc4_key = md5(fk.salt + aqwerty + syskey + anum).digest()
            samkey_data = ARC4.new(rc4_key).encrypt(fk.key + fk.checksum)
            samkey = samkey_data[:16]
            checksum = samkey_data[16:]

            if checksum != md5(samkey + anum + samkey + aqwerty).digest():
                raise ValueError("SAM key checksum validation failed!")
            return samkey

        # SAM key revision 0x02 involving AES  (samsrv.dll: KEDecryptKeyWithAES)
        fk = c_sam.SAM_KEY_AES(f_key)
        key_data = f_key[len(c_sam.SAM_KEY_AES) : len(c_sam.SAM_KEY_AES) + fk.data_len]
        checksum_data = f_key[
            len(c_sam.SAM_KEY_AES) + fk.data_len : len(c_sam.SAM_KEY_AES) + fk.data_len + fk.checksum_len
        ]
        samkey = AES.new(syskey, AES.MODE_CBC, fk.salt).decrypt(key_data)[:16]
        checksum = AES.new(syskey, AES.MODE_CBC, fk.salt).decrypt(checksum_data)[:32]

        if checksum != sha256(samkey).digest():
            raise ValueError("SAM key checksum validation failed!")
        return samkey

    @export(record=SamRecord)
    def sam(self) -> Iterator[SamRecord]:
        """Dump SAM entries

        The Security Account Manager (SAM) registry hive contains registry keys that store usernames, full names and
        passwords in a hashed format, either an LM or NT hash.

        References:
            - https://en.wikipedia.org/wiki/Security_Account_Manager

        Yields SamRecords with fields:

        .. code-block:: text

            ts (datetime): The creation date.
            rid (uint32): The RID.
            fullname (string): Parsed fullname.
            username (string): Parsed username.
            admincomment (string): Parsed admin comment.
            usercomment (string): Parsed user comment.
            lastlogin (datetime): Parsed last login date.
            lastpasswordset (datetime): Parsed last password set date.
            lastincorrectlogin (datetime): Parsed last incorrect login date.
            flags (uint32): Parsed flags.
            countrycode (uint16): Parsed country code (international country calling code).
            failedlogins (uint32): Parsed failed logins, reset after sucessful login.
            logins (uint32): Parsed logins (max 0xFFFF = 65535).
            lm (string): Parsed LM-hash.
            nt (string): Parsed NT-hash.
        """

        syskey = self.target.lsa.syskey  # aka. bootkey
        samkey = self.calculate_samkey(syskey)  # aka. hashed bootkey or hbootkey

        almpassword = b"LMPASSWORD\0"
        antpassword = b"NTPASSWORD\0"

        for users_key in self.target.registry.keys(f"{self.SAM_KEY}\\Users"):
            for user_key in users_key.subkeys():
                if user_key.name == "Names":
                    continue

                f = c_sam.user_F(user_key.value("F").value)
                user_v = user_key.value("V").value
                v = c_sam.user_V(user_v)
                v_data = user_v[0xCC:]

                u_username = v_data[v.username_ofs : v.username_ofs + v.username_len].decode("utf-16-le")
                u_fullname = v_data[v.fullname_ofs : v.fullname_ofs + v.fullname_len].decode("utf-16-le")
                u_admin_comment = v_data[v.admin_comment_ofs : v.admin_comment_ofs + v.admin_comment_len].decode(
                    "utf-16-le"
                )
                u_user_comment = v_data[v.user_comment_ofs : v.user_comment_ofs + v.user_comment_len].decode(
                    "utf-16-le"
                )

                u_lmpw = v_data[v.lmpw_ofs : v.lmpw_ofs + v.lmpw_len]
                u_ntpw = v_data[v.ntpw_ofs : v.ntpw_ofs + v.ntpw_len]

                lm_hash = decrypt_single_hash(f.rid, samkey, u_lmpw, almpassword).hex()
                nt_hash = decrypt_single_hash(f.rid, samkey, u_ntpw, antpassword).hex()

                names_key = self.target.registry.key(f"{self.SAM_KEY}\\Users\\Names\\{u_username}")

                yield SamRecord(
                    ts=names_key.ts,
                    rid=f.rid,
                    fullname=u_fullname,
                    username=u_username,
                    admincomment=u_admin_comment,
                    usercomment=u_user_comment,
                    lastlogin=ts.wintimestamp(f.t_last_login),
                    lastpasswordset=ts.wintimestamp(f.t_last_password_set),
                    lastincorrectlogin=ts.wintimestamp(f.t_last_incorrect_login),
                    flags=f.ACB_bits,
                    countrycode=f.country_code,
                    logins=f.logins,
                    failedlogins=f.failedcnt,
                    lm=lm_hash,
                    nt=nt_hash,
                    _target=self.target,
                )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugin import export
from dissect.target.plugins.os.windows.dpapi.keyprovider.keyprovider import (
    KeyProviderPlugin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator


class EmptyKeyProviderPlugin(KeyProviderPlugin):
    """Empty key provider plugin."""

    __namespace__ = "dpapi.keyprovider.empty"

    def check_compatible(self) -> None:
        return

    @export(output="yield")
    def keys(self) -> Iterator[tuple[str, str]]:
        """Yield an empty string."""
        yield self.__namespace__, ""



from __future__ import annotations

from datetime import datetime
from pathlib import Path
from statistics import median
from typing import TYPE_CHECKING

from dissect.util import ts

from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from datetime import datetime
    from pathlib import Path


class GenericPlugin(Plugin):
    """Generic plugin for Unix targets."""

    def check_compatible(self) -> None:
        pass

    @export(property=True)
    def activity(self) -> datetime | None:
        """Return last seen activity based on filesystem timestamps."""
        var_log = self.target.fs.path("/var/log")
        return calculate_last_activity(var_log)

    @export(property=True)
    def install_date(self) -> datetime | None:
        """Return the likely install date of the operating system."""

        # Although this purports to be a generic function for Unix targets,
        # these paths are Linux specific.
        files = [
            # Debian
            "/var/log/installer/install-journal.txt",
            "/var/log/installer/syslog",
            "/var/lib/dpkg/arch",
            # RedHat
            "/root/anaconda-ks.cfg",
            # Generic
            "/etc/hostname",
            "/etc/machine-id",
        ]
        dates = []

        for f in files:
            p = self.target.fs.path(f)
            if p.exists():
                dates.append(p.stat().st_mtime)

        if dates:
            return ts.from_unix(median(dates))

        # As a fallback if none of the above files are found, we attempt
        # to discover the birth date of the root filesystem.
        #
        # If the change time and modify time of the root dir are equal,
        # this is likely the creation timestamp of the root fs.
        root_stat = self.target.fs.stat("/")
        if root_stat.st_ctime == root_stat.st_mtime:
            return ts.from_unix(root_stat.st_ctime)
        return None


def calculate_last_activity(folder: Path, recursive: bool = False) -> datetime | None:
    if not folder.exists() or not folder.is_dir():
        return None

    last_seen = 0
    for file in folder.rglob("*") if recursive else folder.iterdir():
        if not file.exists():
            continue
        if file.stat().st_mtime > last_seen:
            last_seen = file.stat().st_mtime

    if last_seen != 0:
        return ts.from_unix(last_seen)
    return None



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

import dissect.vmfs as vmfs
from dissect.vmfs.c_vmfs import c_vmfs
from dissect.vmfs.descriptor import FileDescriptor

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.vmfs.vmfs import FileDescriptor


class VmfsFilesystem(Filesystem):
    __type__ = "vmfs"

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.vmfs = vmfs.VMFS(fh)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        """Detect a VMFS filesystem on a given file-like object."""
        fh.seek(c_vmfs.FS3_FS_HEADER_OFFSET)
        sector = fh.read(512)
        return int.from_bytes(sector[:4], "little") in (
            c_vmfs.VMFS_MAGIC_NUMBER,
            c_vmfs.VMFSL_MAGIC_NUMBER,
        )

    def get(self, path: str) -> FilesystemEntry:
        return VmfsFilesystemEntry(self, path, self._get_node(path))

    def _get_node(self, path: str, node: FileDescriptor | None = None) -> FileDescriptor:
        """Returns an internal VMFS entry for a given path and optional relative entry."""
        try:
            return self.vmfs.get(path, node)
        except vmfs.FileNotFoundError as e:
            raise FileNotFoundError(path) from e
        except vmfs.NotADirectoryError as e:
            raise NotADirectoryError(path) from e
        except vmfs.NotASymlinkError as e:
            raise NotASymlinkError(path) from e
        except vmfs.Error as e:
            raise FileNotFoundError(path) from e


class VmfsFilesystemEntry(FilesystemEntry):
    fs: VmfsFilesystem
    entry: FileDescriptor

    def get(self, path: str) -> FilesystemEntry:
        """Get a filesystem entry relative from the current one."""
        full_path = fsutil.join(self.path, path, alt_separator=self.fs.alt_separator)
        return VmfsFilesystemEntry(self.fs, full_path, self.fs._get_node(path, self.entry))

    def open(self) -> BinaryIO:
        """Returns file handle (file-like object)."""
        if self.is_dir():
            raise IsADirectoryError(self.path)
        return self._resolve().entry.open()

    def _iterdir(self) -> Iterator[FileDescriptor]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        if self.is_symlink():
            for f in self.readlink_ext().iterdir():
                yield f
        else:
            for f in self.entry.iterdir():
                if f.name in (".", ".."):
                    continue

                yield f

    def iterdir(self) -> Iterator[str]:
        """List the directory contents of a directory. Returns a generator of strings."""
        for f in self._iterdir():
            yield f.name

    def scandir(self) -> Iterator[FilesystemEntry]:
        """List the directory contents of this directory. Returns a generator of filesystem entries."""
        for f in self._iterdir():
            path = fsutil.join(self.path, f.name, alt_separator=self.fs.alt_separator)
            yield VmfsFilesystemEntry(self.fs, path, f.file_descriptor)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a directory."""
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.is_dir()
        except FilesystemError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a file."""
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.is_file()
        except FilesystemError:
            return False

    def is_symlink(self) -> bool:
        """Return whether this entry is a link."""
        return self.entry.is_symlink()

    def readlink(self) -> str:
        """Read the link of the given path if it is a symlink. Returns a string."""
        if not self.is_symlink():
            raise NotASymlinkError

        return self.entry.link

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        """Return the stat information of this entry."""
        return self._resolve(follow_symlinks=follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        """Return the stat information of the given path, without resolving links."""
        meta = self.entry.metadata

        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        st_info = [
            self.entry.mode,
            self.entry.address,
            id(self.fs),
            meta.linkCount,
            meta.uid,
            meta.gid,
            self.entry.size,
            meta.atime,
            meta.mtime,
            meta.ctime,
        ]

        return fsutil.stat_result(st_info)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.plugins.os.windows.log.mssql import MssqlPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


def test_mssql_errorlog(target_win_users: Target, hive_hklm: VirtualHive, fs_win: Filesystem) -> None:
    errorlog_file = absolute_path("_data/plugins/os/windows/log/mssql/errorlog")
    target_errorlog_name = "/sysvol/Temp/MSSQL/Log/ERRORLOG"

    _, _, map_path = target_errorlog_name.partition("sysvol/")
    fs_win.map_file(map_path, errorlog_file)

    errorlog_name = "SOFTWARE\\Microsoft\\Microsoft SQL Server\\MSSQL69.MyInstance\\SQLServerAgent"
    errorlog_key = VirtualKey(hive_hklm, errorlog_name)
    hive_hklm.map_key(errorlog_name, errorlog_key)
    errorlog_key.add_value(
        "ErrorLogFile", VirtualValue(hive_hklm, "ErrorLogFile", "C:\\Temp\\MSSQL\\Log\\SQLAGENT.OUT")
    )

    datapath_name = "SOFTWARE\\Microsoft\\Microsoft SQL Server\\MSSQL69.MyInstance\\MSSQLServer"
    datapath_key = VirtualKey(hive_hklm, datapath_name)
    hive_hklm.map_key(datapath_name, datapath_key)
    datapath_key.add_value("DefaultData", VirtualValue(hive_hklm, "DefaultData", "C:\\Temp\\MSSQL\\Data"))

    target_win_users.add_plugin(MssqlPlugin)
    records = list(target_win_users.mssql())
    assert len(records) == 101

    record = records[51]
    assert str(record.ts) == "2024-04-08 12:16:41.190000+00:00"
    assert record.instance == "MSSQL69.MyInstance"
    assert record.process == "Server"
    assert record.message.startswith("The SQL Server Network Interface library could not register")
    assert record.path == "C:\\Temp\\MSSQL\\Log\\ERRORLOG"



from __future__ import annotations

import errno
import logging
from ctypes import c_void_p
from functools import lru_cache
from typing import TYPE_CHECKING, BinaryIO

from dissect.util.feature import Feature, feature_enabled

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.filesystem import Filesystem, FilesystemEntry

if feature_enabled(Feature.BETA):
    from fuse3 import FuseOSError, Operations
    from fuse3.c_fuse import fuse_config_p, fuse_conn_info_p

    HAS_FUSE3 = True
else:
    from fuse import FuseOSError, Operations

    fuse_config_p = c_void_p
    fuse_conn_info_p = c_void_p

    HAS_FUSE3 = False


log = logging.getLogger(__name__)

CACHE_SIZE = 64 * 1024  # Tuned for ~0.7GB of memory usage on extFS


class DissectMount(Operations):
    def __init__(self, fs: Filesystem):
        self.fs = fs
        self.file_handles: dict[int, BinaryIO] = {}
        self.dir_handles: dict[int, FilesystemEntry] = {}

        self._get = lru_cache(CACHE_SIZE)(self._get)
        self.getattr = lru_cache(CACHE_SIZE)(self.getattr)

    def _get(self, path: str) -> FilesystemEntry:
        try:
            return self.fs.get(path)
        except Exception:
            raise FuseOSError(errno.ENOENT)

    def init(self, path: str, conn: fuse_conn_info_p | None = None, cfg: fuse_config_p | None = None) -> None:
        if cfg:
            # Enables the use of inodes in getattr
            cfg.contents.use_ino = 1

    def getattr(self, path: str, fh: int | None = None) -> dict:
        fe = self._get(path)

        try:
            st = fe.lstat()

            return {
                key: getattr(st, key)
                for key in (
                    "st_atime",
                    "st_ctime",
                    "st_ino",
                    "st_gid",
                    "st_mode",
                    "st_mtime",
                    "st_nlink",
                    "st_size",
                    "st_uid",
                )
            }

        except Exception:
            raise FuseOSError(errno.EIO)

    getxattr = None

    listxattr = None

    def open(self, path: str, flags: int) -> int:
        entry = self._get(path)

        try:
            fh = entry.open()
        except Exception:
            raise FuseOSError(errno.ENOENT)

        fno = id(fh)
        self.file_handles[fno] = fh
        return fno

    def opendir(self, path: str) -> int:
        entry = self._get(path)

        fno = id(entry)
        self.dir_handles[fno] = entry
        return fno

    def read(self, path: str, size: int, offset: int, fh: int) -> bytes:
        if fh not in self.file_handles:
            raise FuseOSError(errno.EBADFD)

        fobj = self.file_handles[fh]

        try:
            fobj.seek(offset)
            return fobj.read(size)
        except Exception:
            log.exception("Exception in fuse::read")
            raise FuseOSError(errno.EIO)

    def readdir(self, path: str, fh: int, flags: int = 0) -> Iterator[str]:
        if fh not in self.dir_handles:
            raise FuseOSError(errno.EBADFD)

        fobj = self.dir_handles[fh]

        try:
            yield "."
            yield ".."

            yield from fobj.iterdir()
        except Exception:
            log.exception("Exception in fuse::readdir")
            raise FuseOSError(errno.EIO)

    def readlink(self, path: str) -> str:
        fe = self._get(path)

        try:
            return fe.readlink()
        except Exception:
            raise FuseOSError(errno.EIO)

    def release(self, path: str, fh: int) -> int:
        if file := self.file_handles.get(fh):
            file.close()

        del self.file_handles[fh]
        return 0

    def releasedir(self, path: str, fh: int) -> int:
        del self.dir_handles[fh]
        return 0

    if HAS_FUSE3:
        # Define the fuse3 bindings here

        def lseek(self, path: str, off: int, whence: int, fh: int) -> int:
            if file := self.file_handles.get(fh):
                return file.seek(off, whence)
            raise FuseOSError(errno.EBADFD)



from __future__ import annotations

from struct import unpack
from typing import TYPE_CHECKING

from defusedxml import ElementTree
from dissect.cstruct import cstruct
from dissect.regf.c_regf import (
    REG_BINARY,
    REG_DWORD,
    REG_DWORD_BIG_ENDIAN,
    REG_EXPAND_SZ,
    REG_LINK,
    REG_MULTI_SZ,
    REG_NONE,
    REG_QWORD,
    REG_SZ,
)

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.target import Target

policy_def = """
struct registry_policy_header {
    uint32   signature;
    uint32   version;
};
"""
c_adpolicy = cstruct().load(policy_def)

ADPolicyRecord = TargetRecordDescriptor(
    "windows/adpolicy",
    [
        ("datetime", "last_modification_time"),
        ("datetime", "last_access_time"),
        ("datetime", "creation_time"),
        ("string", "guid"),
        ("string", "key"),
        ("string", "value"),
        ("uint32", "size"),
        ("dynamic", "data"),
        ("path", "path"),
    ],
)


class ADPolicyPlugin(Plugin):
    """Plugin that parses AD policies present on Windows Server and Desktop systems

    References:
        - https://docs.microsoft.com/en-us/previous-versions/windows/desktop/policy/registry-policy-file-format
    """

    def __init__(self, target: Target):
        super().__init__(target)
        self.paths = "windows/sysvol/domain/policies/", "windows/system32/GroupPolicy/DataStore/"
        self.dirs = []

        for fs in self.target.filesystems:
            for path in self.paths:
                self.dirs.append(fs.path(path))

    def check_compatible(self) -> None:
        if len([d for d in self.dirs if d.exists()]) <= 0:
            raise UnsupportedPluginError("No AD policy directories found")

    def _xmltasks(self, policy_dir: Path) -> Iterator[ADPolicyRecord]:
        for task_file in policy_dir.rglob("ScheduledTasks.xml"):
            try:
                task_file_stat = task_file.stat()
                xml = task_file.read_text()
                tree = ElementTree.fromstring(xml)
                for task in tree.findall(".//{*}Task"):
                    # https://github.com/python/cpython/issues/83122
                    if (properties := task.find("Properties")) is None:
                        properties = task

                    task_data = ElementTree.tostring(task)
                    yield ADPolicyRecord(
                        last_modification_time=task_file_stat.st_mtime,
                        last_access_time=task_file_stat.st_atime,
                        creation_time=task_file_stat.st_ctime,
                        guid=task.attrib.get("uid"),
                        key="XML",
                        value=properties.attrib.get("appName", None),
                        size=len(task_data),
                        data=task_data,
                        path=task_file,
                        _target=self.target,
                    )
            except Exception as e:  # noqa: PERF203
                self.target.log.warning("Unable to read XML policy file: %s", e)

    @export(record=ADPolicyRecord)
    def adpolicy(self) -> Iterator[ADPolicyRecord]:
        """Return all AD policies (also known as GPOs or Group Policy Objects).

        An Active Directory (AD) maintains global policies that should be adhered by all systems in the domain.
        Example policies are password policies, account lockout policies and group policies.

        References:
            - https://www.windows-active-directory.com/windows-active-directory-policies.html
        """
        for policy_dir in self.dirs:
            if not policy_dir.exists():
                continue

            yield from self._xmltasks(policy_dir)

            # The body consists of registry values in the following format.
            # [key;value;type;size;data]
            for policy_file in policy_dir.rglob("*.pol"):
                policy_file_stat = policy_file.stat()
                if policy_file_stat.st_size <= 8:  # skip empty registry.pol files (only header defined)
                    continue

                fh = policy_file.open()
                policy_header = c_adpolicy.registry_policy_header(fh)
                if policy_header.signature != 0x67655250:
                    self.target.log.warning("Invalid Registry.pol header encountered in file %s", policy_file)
                    continue

                policy_body = fh.read()
                policy_body = policy_body.split(b"]\x00[\x00")

                for policy_line in policy_body:
                    policy_line = policy_line.replace(b"[\x00", b"").replace(b"]\x00", b"")
                    values = policy_line.split(b";\x00", maxsplit=4)

                    if len(values) == 4:
                        # policies without data will not split on ;\x00'
                        # thus we have to split the remainder variable on just ';'
                        policy_reg_key, policy_reg_value, policy_reg_type, remainder = values
                        policy_reg_data_size, policy_reg_data = remainder.split(b";")
                    else:
                        (
                            policy_reg_key,
                            policy_reg_value,
                            policy_reg_type,
                            policy_reg_data_size,
                            policy_reg_data,
                        ) = values

                    policy_reg_data_size = unpack("i", policy_reg_data_size)[0]
                    policy_reg_type = unpack("i", policy_reg_type)[0]

                    policy_reg_data = _decode_policy_reg_data(policy_reg_type, policy_reg_data[:policy_reg_data_size])

                    yield ADPolicyRecord(
                        last_modification_time=policy_file_stat.st_mtime,
                        last_access_time=policy_file_stat.st_atime,
                        creation_time=policy_file_stat.st_ctime,
                        guid=policy_file.parts[4],
                        key=policy_reg_key.decode("utf-16-le").rstrip("\x00"),
                        value=policy_reg_value.decode("utf-16-le").rstrip("\x00"),
                        size=policy_reg_data_size,
                        data=policy_reg_data,
                        path=policy_file,
                        _target=self.target,
                    )


def _decode_policy_reg_data(policy_reg_type: int, policy_reg_data: bytes) -> bytes | str | int | None:
    if policy_reg_data is None or policy_reg_type == REG_NONE:
        return policy_reg_data
    if policy_reg_type in (REG_EXPAND_SZ, REG_SZ, REG_MULTI_SZ, REG_LINK):
        # REG_SZ, REG_MULTI_SZ, and REG_EXPAND_SZ types get a null terminating character added. We remove that here.
        # ref: https://docs.microsoft.com/en-us/windows/win32/sysinfo/registry-value-types
        return policy_reg_data.decode("utf-16-le").rstrip("\x00")
    if policy_reg_type == REG_DWORD:
        return unpack("i", policy_reg_data)[0]
    if policy_reg_type == REG_DWORD_BIG_ENDIAN:
        return unpack(">i", policy_reg_data)[0]
    if policy_reg_type == REG_QWORD:
        return unpack("q", policy_reg_data)[0]
    if policy_reg_type == REG_BINARY:
        return policy_reg_data
    return None



from __future__ import annotations

import sys
from unittest.mock import MagicMock

import pytest

from dissect.target.exceptions import NotADirectoryError


def test_smb_filesystem_windows(monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        if "dissect.target.filesystems.smb" in sys.modules:
            m.delitem(sys.modules, "dissect.target.filesystems.smb")

        mock_impacket = MagicMock()
        m.setitem(sys.modules, "impacket", mock_impacket)
        m.setitem(sys.modules, "impacket.nt_errors", mock_impacket.nt_errors)
        m.setitem(sys.modules, "impacket.smb", mock_impacket.smb)
        m.setitem(sys.modules, "impacket.smb3structs", mock_impacket.smb3structs)
        m.setitem(sys.modules, "impacket.smbconnection", mock_impacket.smbconnection)

        from dissect.target.filesystems.smb import SmbFilesystem

        mock_dir = MagicMock()
        mock_dir.get_longname.return_value = "testdir"
        mock_dir.is_directory.return_value = True

        mock_conn = MagicMock()
        mock_conn.listPath.return_value = [mock_dir]

        fs = SmbFilesystem(mock_conn, "C$")

        root = fs.get("")
        mock_conn.listPath.assert_not_called()

        assert root.path == ""
        assert root.is_dir()
        assert not root.is_file()
        assert not root.is_symlink()

        entry = fs.get("testdir")
        mock_conn.listPath.assert_called_with("C$", "testdir")

        assert entry.path == "testdir"
        assert entry.name == "testdir"
        assert entry.is_dir()
        assert not entry.is_file()
        assert not entry.is_symlink()

        mock_file = MagicMock()
        mock_file.get_longname.return_value = "testfile.txt"
        mock_file.is_directory.return_value = False
        mock_file.get_filesize.return_value = 69
        mock_file.get_atime_epoch.return_value = 1
        mock_file.get_mtime_epoch.return_value = 2
        mock_file.get_ctime_epoch.return_value = 3

        mock_conn.listPath.return_value = [mock_file]

        entries = list(entry.scandir())
        mock_conn.listPath.assert_called_with("C$", "testdir/*")

        assert len(entries) == 1
        assert entries[0].path == "testdir/testfile.txt"
        assert entries[0].name == "testfile.txt"
        assert not entries[0].is_dir()
        assert entries[0].is_file()
        assert not entries[0].is_symlink()

        with pytest.raises(NotADirectoryError):
            entries[0].listdir()

        entries[0].open()
        mock_conn.connectTree.assert_called_with("C$")
        mock_conn.openFile.assert_called_once()

        stat_result = entries[0].stat()
        assert stat_result.st_mode == 0o100755
        assert stat_result.st_ino == 3086484436
        assert stat_result.st_atime == 1.0
        assert stat_result.st_mtime == 2.0
        assert stat_result.st_ctime == 3.0



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.os.windows import adpolicy
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_adpolicy_plugin(target_win: Target, fs_win: VirtualFilesystem) -> None:
    policy_dir = absolute_path("_data/plugins/os/windows/adpolicy/")

    fs_win.map_dir("Windows/sysvol/domain/policies", policy_dir)

    target_win.add_plugin(adpolicy.ADPolicyPlugin)

    records = list(target_win.adpolicy())
    assert len(records) == 10



from __future__ import annotations

import io
from typing import TYPE_CHECKING

from dissect.target.helpers import fsutil
from dissect.target.helpers.regutil import VirtualHive, VirtualKey
from dissect.target.plugins.apps.productivity.msoffice import MSOffice
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_office_startup_default_machine(
    target_win_users: Target, fs_win: VirtualFilesystem, hive_hklm: VirtualHive
) -> None:
    """Test if machine-scoped startup items are found in default locations."""

    office_install_path = "C:/Office"
    key_path = "Software\\Microsoft\\Office\\16.0\\Word\\InstallRoot"
    startup_item_path = fsutil.join(office_install_path, "STARTUP/plugin.wll")

    install_root_key = VirtualKey(hive_hklm, key_path)
    install_root_key.add_value("Path", office_install_path)
    hive_hklm.map_key(key_path, install_root_key)
    fs_win.map_file_fh(startup_item_path.removeprefix("C:/"), io.BytesIO(b"Plugin Data"))

    office_plugin = MSOffice(target_win_users)
    startup_items = list(office_plugin.startup())

    assert len(startup_items) == 1
    item, *_ = startup_items
    assert item.path == startup_item_path
    assert item.creation_time == target_win_users.fs.stat(startup_item_path).st_birthtime


def test_office_startup_default_user(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    """Test if user-scoped startup items are found in default locations."""

    startup_item_path = "C:/Users/John/AppData/Roaming/Microsoft/Templates/normal.dotx"
    fs_win.map_file_fh(startup_item_path.removeprefix("C:/"), io.BytesIO(b"Template Data"))

    office_plugin = MSOffice(target_win_users)
    startup_items = list(office_plugin.startup())

    assert len(startup_items) == 1
    item, *_ = startup_items
    assert item.path == startup_item_path
    assert item.creation_time == target_win_users.fs.stat(startup_item_path).st_birthtime


def test_office_startup_options(target_win_users: Target, fs_win: VirtualFilesystem, hive_hklm: VirtualHive) -> None:
    """Test if startup items are found in custom specified locations."""

    startup_item_path = "C:/FUNWAREZ/innocent.dll"
    key_path = "Software\\Microsoft\\Office\\16.0\\Word\\Options"

    custom_startup_key = VirtualKey(hive_hklm, key_path)
    custom_startup_key.add_value("STARTUP-PATH", "C:/FUNWAREZ")
    hive_hklm.map_key(key_path, custom_startup_key)
    fs_win.map_file_fh(startup_item_path.removeprefix("C:/"), io.BytesIO(b"Exploit Data"))

    office_plugin = MSOffice(target_win_users)
    startup_items = list(office_plugin.startup())

    assert len(startup_items) == 1
    item, *_ = startup_items
    assert item.path == startup_item_path
    assert item.creation_time == target_win_users.fs.stat(startup_item_path).st_birthtime


def test_office_com_addin(target_win_users: Target, hive_hklm: VirtualHive) -> None:
    """Test if COM add-ins are found."""

    addin_prog_id = "ExcelAddin"
    addin_key_path = f"Software\\Microsoft\\Office\\Excel\\Addins\\{addin_prog_id}"
    cls_id = "{ADC6CB82-424C-11D2-952A-00C04FA34F05}"

    addin_key = VirtualKey(hive_hklm, addin_key_path)
    addin_key.add_value("FriendlyName", "An Excel com addin")
    addin_key.add_value("LoadBehavior", 3)
    hive_hklm.map_key(addin_key_path, addin_key)
    hive_hklm.map_value(f"Software\\Classes\\{addin_prog_id}\\CLSID", "(Default)", cls_id)
    hive_hklm.map_value(f"Software\\Classes\\CLSID\\{cls_id}\\InprocServer32", "(Default)", "c:\\payload.exe")

    office_plugin = MSOffice(target_win_users)
    startup_items = list(office_plugin.native())

    assert len(startup_items) == 1
    item, *_ = startup_items
    assert item.name == "An Excel com addin"
    assert item.type == "com"
    assert item.load_behavior == "Autostart"
    assert item.codebases == ["c:\\payload.exe"]


def test_office_vsto_addin(target_win_users: Target, fs_win: VirtualFilesystem, hive_hklm: VirtualHive) -> None:
    """Test if vsto add-ins are found."""

    addin_prog_id = "ExcelAddin"
    addin_key_path = f"Software\\Microsoft\\Office\\Excel\\Addins\\{addin_prog_id}"

    fs_win.map_dir("vsto", absolute_path("_data/plugins/apps/productivity/vsto"))
    addin_key = VirtualKey(hive_hklm, addin_key_path)
    addin_key.add_value("FriendlyName", "An Excel vsto addin")
    addin_key.add_value("LoadBehavior", 2)
    addin_key.add_value("Manifest", "c:\\vsto\\MicrosoftDataStreamerforExcel.vsto|vstolocal")
    hive_hklm.map_key(addin_key_path, addin_key)

    office_plugin = MSOffice(target_win_users)
    startup_items = list(office_plugin.native())

    assert len(startup_items) == 1
    item, *_ = startup_items
    assert item.name == "An Excel vsto addin"
    assert item.type == "vsto"
    assert item.load_behavior == "Autostart"
    assert not item.loaded
    assert item.manifest == "c:\\vsto\\MicrosoftDataStreamerforExcel.vsto|vstolocal"
    assert item.codebases == ["c:\\vsto\\MicrosoftDataStreamerforExcel.dll"]


def test_office_web_addin(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    """Test if web add-ins are found."""

    fs_win.map_dir(
        "users/John/AppData/local/Microsoft/Office/16.0/Wef", absolute_path("_data/plugins/apps/productivity/wef")
    )

    office_plugin = MSOffice(target_win_users)
    startup_items = list(office_plugin.web())

    assert len(startup_items) == 1
    item, *_ = startup_items
    assert item.name == "ChatGPT for MS Word"
    assert item.provider_name == "Apps Do Wonders LLC"
    assert (
        item.manifest == "C:\\Users\\John\\AppData\\local\\Microsoft\\Office\\16.0\\Wef\\Manifests\\wa200007708_1.0.0.0"
    )
    assert item.source_locations == ["https://word-addin.appsdowonders.com/taskpane.html"]
    assert item.version == "1.0.0.0"
